import contextlib
import ssl
from unittest.mock import Mock

import pytest
import trustme


@pytest.fixture
def mock_cache(mocker):
    parent = Mock(
        cache_file=Mock(return_value='path/to/cache.file'),
        from_cache=Mock(return_value=None),
        to_cache=Mock(return_value=None),
    )
    mocker.patch('upsies.utils.http._cache_file', parent.cache_file)
    mocker.patch('upsies.utils.http._from_cache', parent.from_cache)
    mocker.patch('upsies.utils.http._to_cache', parent.to_cache)
    yield parent


@pytest.fixture
def custom_host(ssl_context):
    @contextlib.contextmanager
    def custom_host(httpserver, host, tls=False, port=None):
        httpserver.stop()
        original_host, original_port = httpserver.host, httpserver.port
        httpserver.host = host
        if port:
            httpserver.port = port
        if tls:
            httpserver.ssl_context = ssl_context

        httpserver.start()

        try:
            yield
        finally:
            httpserver.stop()
            httpserver.host, httpserver.port = original_host, original_port
            httpserver.ssl_context = None
            httpserver.start()

    return custom_host

@pytest.fixture(scope='module')
def ssl_context(ca):
    context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
    localhost_cert = ca.issue_cert('localhost')
    localhost_cert.configure_cert(context)
    return context

@pytest.fixture(scope='module')
def ca():
    return trustme.CA()



import os
import re
from unittest.mock import Mock, PropertyMock, call

import pytest

from upsies import __project_name__, errors, utils
from upsies.trackers.mtv import MtvTracker, MtvTrackerConfig, MtvTrackerJobs


@pytest.fixture
def make_tracker():
    def make_tracker(**kwargs):
        options = {
            'username': 'bunny',
            'password': 'hunter2',
            'base_url': 'http://mtv.local',
        }
        options.update(kwargs)
        return MtvTracker(options=options)

    return make_tracker


def test_name_attribute():
    assert MtvTracker.name == 'mtv'


def test_label_attribute():
    assert MtvTracker.label == 'MTV'


def test_TrackerConfig_attribute():
    assert MtvTracker.TrackerConfig is MtvTrackerConfig


def test_MtvTracker_attribute():
    assert MtvTracker.TrackerJobs is MtvTrackerJobs


def test_MtvTracker_base_url_attribute(make_tracker):
    tracker = make_tracker(base_url='http://foo.local')
    assert tracker._base_url == 'http://foo.local'


def test_MtvTracker_login_url_attribute(make_tracker):
    tracker = make_tracker(base_url='http://foo.local')
    assert tracker._login_url == 'http://foo.local/login'


def test_MtvTracker_logout_url_attribute(make_tracker):
    tracker = make_tracker(base_url='http://foo.local')
    assert tracker._logout_url == 'http://foo.local/logout'


def test_MtvTracker_upload_url_attribute(make_tracker):
    tracker = make_tracker(base_url='http://foo.local')
    assert tracker._upload_url == 'http://foo.local/upload.php'


def test_MtvTracker_torrents_url_attribute(make_tracker):
    tracker = make_tracker(base_url='http://foo.local')
    assert tracker._torrents_url == 'http://foo.local/torrents.php'


@pytest.fixture
def tracker_for_login(make_tracker, mocker):
    tracker = make_tracker()
    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(tracker, '_get_token', return_value='mock token'), '_get_token')
    mocks.attach_mock(mocker.patch('upsies.trackers.mtv.tracker.utils.http.post'), 'post')
    mocks.attach_mock(mocker.patch('upsies.trackers.mtv.tracker.utils.html.parse'), 'parse')
    mocks.attach_mock(mocker.patch.object(tracker, '_confirm_login'), '_confirm_login')
    return tracker, mocks

@pytest.mark.asyncio
async def test_MtvTracker_login_without_username(tracker_for_login):
    tracker, mocks = tracker_for_login
    del tracker.options['username']
    with pytest.raises(errors.RequestError, match=r'^Login failed: No username configured$'):
        await tracker._login()
    assert mocks.mock_calls == []

@pytest.mark.asyncio
async def test_MtvTracker_login_without_password(tracker_for_login):
    tracker, mocks = tracker_for_login
    del tracker.options['password']
    with pytest.raises(errors.RequestError, match=r'^Login failed: No password configured$'):
        await tracker._login()
    assert mocks.mock_calls == []

@pytest.mark.asyncio
async def test_MtvTracker_login_succeeds(tracker_for_login):
    tracker, mocks = tracker_for_login
    await tracker._login()
    assert mocks.mock_calls == [
        call._get_token(),
        call.post(
            url=tracker._login_url,
            user_agent=True,
            data={
                'token': mocks._get_token.return_value,
                'username': tracker.options['username'],
                'password': tracker.options['password'],
                'cinfo': '1280|720|24|0',
                'iplocked': '1',
                'submit': 'login',
            },
        ),
        call.parse(mocks.post.return_value),
        call._confirm_login(mocks.parse.return_value),
    ]


@pytest.fixture
def make_tracker_for_logout(make_tracker, mocker):
    def make_tracker_for_logout():
        tracker = make_tracker()
        mocks = Mock()
        mocks.attach_mock(mocker.patch('upsies.trackers.mtv.tracker.utils.http.post'), 'post')
        mocks.attach_mock(mocker.patch.object(tracker, '_get_token'), '_get_token')
        return tracker, mocks

    return make_tracker_for_logout

@pytest.mark.asyncio
async def test_MtvTracker_logout_fails(make_tracker_for_logout):
    tracker, mocks = make_tracker_for_logout()
    exception = errors.RequestError('nope')
    mocks.post.side_effect = exception

    with pytest.raises(type(exception), match=rf'^{re.escape(str(exception))}$'):
        await tracker._logout()

    exp_mock_calls = [
        call._get_token(),
        call.post(
            url=tracker._logout_url,
            data={'token': mocks._get_token.return_value},
            user_agent=True,
        ),
    ]
    assert mocks.mock_calls == exp_mock_calls

@pytest.mark.asyncio
async def test_MtvTracker_logout_succeeds(make_tracker_for_logout):
    tracker, mocks = make_tracker_for_logout()

    await tracker._logout()

    exp_mock_calls = [
        call._get_token(),
        call.post(
            url=tracker._logout_url,
            data={'token': mocks._get_token.return_value},
            user_agent=True,
        ),
    ]
    assert mocks.mock_calls == exp_mock_calls


@pytest.fixture
def make_tracker_for_get_announce_url(make_tracker, mocker):
    def make_tracker_for_get_announce_url():
        tracker = make_tracker()
        mocks = Mock()
        mocks.attach_mock(mocker.patch('upsies.trackers.mtv.tracker.utils.http.get'), 'get')
        mocks.attach_mock(mocker.patch.object(tracker, 'login'), 'login')
        mocks.attach_mock(mocker.patch.object(tracker, 'logout'), 'logout')
        return tracker, mocks

    return make_tracker_for_get_announce_url

@pytest.mark.parametrize(
    argnames='config_url, website_url, exp_url',
    argvalues=(
        ('http://config.url/announce', 'http://website.url/announce', 'http://config.url/announce'),
        ('', 'http://website.url/announce', 'http://website.url/announce'),
        (None, 'http://website.url/announce', 'http://website.url/announce'),
    ),
)
@pytest.mark.asyncio
async def test_MtvTracker_get_announce_url_from_config(config_url, website_url, exp_url, make_tracker_for_get_announce_url, mocker):
    tracker, mocks = make_tracker_for_get_announce_url()
    mocks.get.return_value = f'<html><input value="{website_url}" /></html>'
    if config_url is not None:
        tracker.options['announce_url'] = config_url
    announce_url = await tracker.get_announce_url()
    assert announce_url == exp_url

@pytest.mark.parametrize(
    argnames='html, exp_result',
    argvalues=(
        (
            '<html><input value="http://foo.local:123/announce" /></html>',
            'http://foo.local:123/announce',
        ),
        (
            '<html><input value="https://foo.local:123/announce?d34db33f" /></html>',
            'https://foo.local:123/announce?d34db33f',
        ),
        (
            '<html><input value="https://foo.local:123/announce/d34db33f" /></html>',
            'https://foo.local:123/announce/d34db33f',
        ),
        (
            '<html><input value="https://foo.local:123/d34db33f/announce" /></html>',
            'https://foo.local:123/d34db33f/announce',
        ),
        (
            '<html><input value="https://foo.local:123/anon" /></html>',
            errors.RequestError(
                'Failed to find announce URL - set it manually: '
                f'{__project_name__} set trackers.mtv.announce_url <YOUR URL>',
            ),
        ),
        (
            '<html><input value="" /></html>',
            errors.RequestError(
                'Failed to find announce URL - set it manually: '
                f'{__project_name__} set trackers.mtv.announce_url <YOUR URL>',
            ),
        ),
        (
            '<html><input /></html>',
            errors.RequestError(
                'Failed to find announce URL - set it manually: '
                f'{__project_name__} set trackers.mtv.announce_url <YOUR URL>',
            ),
        ),
        (
            '<html><div>hello</div></html>',
            errors.RequestError(
                'Failed to find announce URL - set it manually: '
                f'{__project_name__} set trackers.mtv.announce_url <YOUR URL>',
            ),
        ),
    ),
)
@pytest.mark.asyncio
async def test_MtvTracker_get_announce_url_from_website(html, exp_result, make_tracker_for_get_announce_url, mocker):
    tracker, mocks = make_tracker_for_get_announce_url()
    tracker.options['announce_url'] = None
    mocks.get.return_value = html

    if isinstance(exp_result, BaseException):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            await tracker.get_announce_url()
    else:
        return_value = await tracker.get_announce_url()
        assert return_value == exp_result

    assert mocks.mock_calls == [
        call.login(),
        call.get(tracker._upload_url, user_agent=True),
        # call.logout(),
    ]


def test_MtvTracker_calculate_piece_size_min_max():
    piece_size_min, piece_size_max = MtvTracker.calculate_piece_size_min_max(123456)
    assert piece_size_min == 32 * 1024
    assert piece_size_max == 8 * 1024 * 1024


@pytest.fixture
def make_tracker_for_upload(make_tracker, mocker, tmp_path):
    def make_tracker_for_upload():
        tracker = make_tracker()
        mocks = Mock()
        mocks.attach_mock(mocker.patch.object(tracker, '_make_autofill_request'), '_make_autofill_request')
        mocks.attach_mock(mocker.patch.object(tracker, '_make_upload_request'), '_make_upload_request')
        return tracker, mocks

    return make_tracker_for_upload

@pytest.mark.asyncio
async def test_MtvTracker_upload(make_tracker_for_upload, mocker):
    tracker, mocks = make_tracker_for_upload()
    tracker_jobs = Mock()

    tracker_page_url = await tracker.upload(tracker_jobs)
    assert tracker_page_url is tracker._make_upload_request.return_value

    assert mocks.mock_calls == [
        call._make_autofill_request(tracker_jobs),
        call._make_upload_request(tracker_jobs, mocks._make_autofill_request.return_value),
    ]


@pytest.fixture
def make_tracker_for_make_autofill_request(make_tracker, mocker, tmp_path):
    def make_tracker_for_make_autofill_request():
        tracker = make_tracker()
        mocks = Mock()
        mocks.attach_mock(mocker.patch.object(tracker, '_prepare_post_data', return_value={'prepared': 'data'}), '_prepare_post_data')
        mocks.attach_mock(mocker.patch('upsies.trackers.mtv.tracker.utils.http.post'), 'post')
        mocks.attach_mock(mocker.patch('upsies.trackers.mtv.tracker.utils.html.parse'), 'parse')
        mocks.attach_mock(mocker.patch.object(tracker, '_raise_error_dupes'), '_raise_error_dupes')
        mocks.attach_mock(mocker.patch.object(tracker, '_raise_error'), '_raise_error')
        mocks.attach_mock(mocker.patch.object(tracker, '_get_form_value'), '_get_form_value')
        return tracker, mocks

    return make_tracker_for_make_autofill_request

@pytest.mark.asyncio
async def test_MtvTracker_make_autofill_request_succeeds(make_tracker_for_make_autofill_request, mocker):
    tracker, mocks = make_tracker_for_make_autofill_request()
    tracker_jobs = Mock(
        post_data_autofill={'autofill': 'data'},
        torrent_filepath='mock/path/to.torrent',
    )
    mocks.post.return_value = '<html><form>autofilled values</form></html>'
    mocks._get_form_value.side_effect = (
        'mock tempfileid',
        'mock tempfilename',
        'mock taglist',
    )

    autofill_post_data = await tracker._make_autofill_request(tracker_jobs)
    assert autofill_post_data == {
        'tempfileid': 'mock tempfileid',
        'tempfilename': 'mock tempfilename',
        'taglist': 'mock taglist',
    }

    assert mocks.mock_calls == [
        call._prepare_post_data(tracker_jobs.post_data_autofill),
        call.post(
            url=tracker._upload_url,
            cache=False,
            user_agent=True,
            follow_redirects=False,
            data={'prepared': 'data'},
            files={
                'file_input': {
                    'file': tracker_jobs.torrent_filepath,
                    'mimetype': 'application/x-bittorrent',
                },
            },
        ),
        call.parse(mocks.post.return_value),
        call._get_form_value(mocks.parse.return_value, 'input', attrs={'name': 'tempfileid'}),
        call._get_form_value(mocks.parse.return_value, 'input', attrs={'name': 'tempfilename'}),
        call._get_form_value(mocks.parse.return_value, 'textarea', attrs={'name': 'taglist'}),
    ]


@pytest.mark.parametrize(
    argnames='unfound_form_field',
    argvalues=(
        (('input',), {'attrs': {'name': 'tempfileid'}}),
        (('input',), {'attrs': {'name': 'tempfilename'}}),
        (('textarea',), {'attrs': {'name': 'taglist'}}),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test_MtvTracker_make_autofill_request_fails(unfound_form_field, make_tracker_for_make_autofill_request, mocker):
    tracker, mocks = make_tracker_for_make_autofill_request()
    tracker_jobs = Mock(
        post_data_autofill={'autofill': 'data'},
        torrent_filepath='mock/path/to.torrent',
    )
    mocks.post.return_value = '<html><form>autofilled values</form></html>'

    def get_form_value_side_effect(doc, *args, **kwargs):
        if (args, kwargs) == unfound_form_field:
            raise ValueError(f'No such form field: ({args}, {kwargs})')

    mocks._get_form_value.side_effect = get_form_value_side_effect

    mocks._raise_error.side_effect = errors.RequestError('Autofilling failed')

    exp_exception = mocks._raise_error.side_effect
    with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
        await tracker._make_autofill_request(tracker_jobs)

    exp_mock_calls = [
        call._prepare_post_data(tracker_jobs.post_data_autofill),
        call.post(
            url=tracker._upload_url,
            cache=False,
            user_agent=True,
            follow_redirects=False,
            data={'prepared': 'data'},
            files={
                'file_input': {
                    'file': tracker_jobs.torrent_filepath,
                    'mimetype': 'application/x-bittorrent',
                },
            },
        ),
        call.parse(mocks.post.return_value),
    ]

    # We don't care about the order of _get_form_value() calls
    assert mocks.mock_calls[:len(exp_mock_calls)] == exp_mock_calls
    # TODO: Figure out how to assert that every mocks.mock_calls[len(exp_mock_calls):-1]
    #       is call._get_from_value() without specifying the exact arguments.
    assert mocks.mock_calls[-1] == call._raise_error(mocks.parse.return_value, msg_prefix='Upload failed')


@pytest.fixture
def make_tracker_for_make_upload_request(make_tracker, mocker, tmp_path):
    def make_tracker_for_make_upload_request():
        tracker = make_tracker()
        mocks = Mock()
        mocks.attach_mock(mocker.patch.object(tracker, '_prepare_post_data', return_value={'prepared': 'data'}), '_prepare_post_data')
        mocks.attach_mock(mocker.patch('upsies.trackers.mtv.tracker.utils.http.post'), 'post')
        mocks.attach_mock(mocker.patch('upsies.trackers.mtv.tracker.utils.html.parse'), 'parse')
        mocks.attach_mock(mocker.patch.object(tracker, '_raise_error_dupes'), '_raise_error_dupes')
        mocks.attach_mock(mocker.patch.object(tracker, '_raise_error'), '_raise_error')
        return tracker, mocks

    cwd_orig = os.getcwd()
    os.chdir(tmp_path)
    try:
        yield make_tracker_for_make_upload_request
    finally:
        os.chdir(cwd_orig)

@pytest.mark.asyncio
async def test_MtvTracker_make_upload_request_succeeds(make_tracker_for_make_upload_request, mocker):
    tracker, mocks = make_tracker_for_make_upload_request()
    tracker_jobs = Mock(post_data_upload={'upload': 'data'})
    autofill_post_data = {'autofill': 'data'}
    exp_post_data = {
        'prepared': 'data',
        'autofill': 'data',
    }
    mocks.post.return_value = Mock(headers={'location': '/torrents.php?id=123'})

    torrent_page_url = await tracker._make_upload_request(tracker_jobs, autofill_post_data)
    assert torrent_page_url == f'{tracker._base_url}/torrents.php?id=123'

    assert mocks.mock_calls == [
        call._prepare_post_data(tracker_jobs.post_data_upload),
        call.post(
            url=tracker._upload_url,
            cache=False,
            user_agent=True,
            follow_redirects=False,
            data=exp_post_data,
        ),
    ]

@pytest.mark.asyncio
async def test_MtvTracker_make_upload_request_finds_dupes(make_tracker_for_make_upload_request, mocker):
    tracker, mocks = make_tracker_for_make_upload_request()
    tracker_jobs = Mock(post_data_upload={'upload': 'data'})
    autofill_post_data = {'autofill': 'data'}
    exp_post_data = {
        'prepared': 'data',
        'autofill': 'data',
    }
    mocks._raise_error_dupes.side_effect = errors.FoundDupeError(['dupe file'])
    mocks.post.return_value = Mock(headers={})

    exp_exception = mocks._raise_error_dupes.side_effect
    with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
        await tracker._make_upload_request(tracker_jobs, autofill_post_data)

    assert mocks.mock_calls == [
        call._prepare_post_data(tracker_jobs.post_data_upload),
        call.post(
            url=tracker._upload_url,
            cache=False,
            user_agent=True,
            follow_redirects=False,
            data=exp_post_data,
        ),
        call.parse(mocks.post.return_value),
        call._raise_error_dupes(mocks.parse.return_value),
    ]

@pytest.mark.asyncio
async def test_MtvTracker_make_upload_request_finds_error_message(make_tracker_for_make_upload_request, mocker):
    tracker, mocks = make_tracker_for_make_upload_request()
    tracker_jobs = Mock(post_data_upload={'upload': 'data'})
    autofill_post_data = {'autofill': 'data'}
    exp_post_data = {
        'prepared': 'data',
        'autofill': 'data',
    }
    mocks._raise_error.side_effect = errors.RequestError('Your upload sucks monkey balls.')
    mocks.post.return_value = Mock(headers={})

    exp_exception = mocks._raise_error.side_effect
    with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
        await tracker._make_upload_request(tracker_jobs, autofill_post_data)

    assert mocks.mock_calls == [
        call._prepare_post_data(tracker_jobs.post_data_upload),
        call.post(
            url=tracker._upload_url,
            cache=False,
            user_agent=True,
            follow_redirects=False,
            data=exp_post_data,
        ),
        call.parse(mocks.post.return_value),
        call._raise_error_dupes(mocks.parse.return_value),
        call._raise_error(mocks.parse.return_value, msg_prefix='Upload failed'),
    ]

@pytest.mark.asyncio
async def test_MtvTracker_make_upload_request_cannot_find_error_message(make_tracker_for_make_upload_request, mocker, tmp_path):
    tracker, mocks = make_tracker_for_make_upload_request()
    tracker_jobs = Mock(
        post_data_upload={'upload': 'data'},
        content_path='path/to/Foo.2012-HEYHO/',
    )
    autofill_post_data = {'autofill': 'data'}
    exp_post_data = {
        'prepared': 'data',
        'autofill': 'data',
    }
    mocks._raise_error.side_effect = RuntimeError('My error discovery sucks monkey balls.')
    mocks.post.return_value = Mock(headers={
        'mock': 'header',
        'fali': 'fala',
    })

    exp_exception = mocks._raise_error.side_effect
    with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
        await tracker._make_upload_request(tracker_jobs, autofill_post_data)

    assert mocks.mock_calls == [
        call._prepare_post_data(tracker_jobs.post_data_upload),
        call.post(
            url=tracker._upload_url,
            cache=False,
            user_agent=True,
            follow_redirects=False,
            data=exp_post_data,
        ),
        call.parse(mocks.post.return_value),
        call._raise_error_dupes(mocks.parse.return_value),
        call._raise_error(mocks.parse.return_value, msg_prefix='Upload failed'),
    ]

    dumped_headers_file = 'Foo.2012-HEYHO.headers'
    dumped_headers = (tmp_path / dumped_headers_file).read_text()
    assert dumped_headers == 'mock: header\nfali: fala\n'


@pytest.fixture
def make_tracker_for_prepare_post_data(make_tracker, mocker):
    def make_tracker_for_prepare_post_data():
        tracker = make_tracker()
        mocks = Mock()
        mocks.attach_mock(mocker.patch.object(tracker, '_get_auth', return_value='myauth'), '_get_auth')
        return tracker, mocks

    return make_tracker_for_prepare_post_data

@pytest.mark.parametrize(
    argnames='post_data, exp_post_data',
    argvalues=(
        (
            {'foo': 'bar', 1: 2, 'nothing': None},
            {'foo': 'bar', '1': '2', 'auth': 'myauth'},
        ),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test_MtvTracker_prepare_post_data(post_data, exp_post_data, make_tracker_for_prepare_post_data, mocker):
    tracker, mocks = make_tracker_for_prepare_post_data()
    return_value = await tracker._prepare_post_data(post_data)
    assert return_value == exp_post_data
    assert mocks.mock_calls == [
        call._get_auth(),
    ]


@pytest.fixture
def make_tracker_for_get_token(make_tracker, mocker):
    def make_tracker_for_get_token():
        tracker = make_tracker()
        mocks = Mock()
        mocks.attach_mock(mocker.patch('upsies.trackers.mtv.tracker.utils.http.get'), 'get')
        mocks.attach_mock(mocker.patch('upsies.trackers.mtv.tracker.utils.html.parse'), 'parse')
        mocks.attach_mock(mocker.patch.object(tracker, '_get_form_value'), '_get_form_value')
        return tracker, mocks

    return make_tracker_for_get_token

@pytest.mark.parametrize(
    argnames='is_logged_in, exp_url_attribute',
    argvalues=(
        (True, '_base_url'),
        (False, '_login_url'),
    ),
)
@pytest.mark.asyncio
async def test_MtvTracker_get_token(is_logged_in, exp_url_attribute, make_tracker_for_get_token, mocker):
    tracker, mocks = make_tracker_for_get_token()
    mocker.patch.object(type(tracker), 'is_logged_in', PropertyMock(return_value=is_logged_in))
    return_value = await tracker._get_token()
    assert return_value is mocks._get_form_value.return_value
    exp_url = getattr(tracker, exp_url_attribute)
    assert mocks.mock_calls == [
        call.get(exp_url, user_agent=True),
        call.parse(mocks.get.return_value),
        call._get_form_value(mocks.parse.return_value, 'input', attrs={'name': 'token'}),
    ]


@pytest.fixture
def make_tracker_for_get_auth(make_tracker, mocker):
    def make_tracker_for_get_auth():
        tracker = make_tracker()
        mocks = Mock()
        mocks.attach_mock(mocker.patch.object(tracker, 'login'), 'login')
        mocks.attach_mock(mocker.patch('upsies.trackers.mtv.tracker.utils.http.get'), 'get')
        mocks.attach_mock(mocker.patch('upsies.trackers.mtv.tracker.utils.html.parse'), 'parse')
        mocks.attach_mock(mocker.patch.object(tracker, '_get_form_value'), '_get_form_value')
        return tracker, mocks

    return make_tracker_for_get_auth

@pytest.mark.asyncio
async def test_MtvTracker_get_auth_with_previously_gotten_auth(make_tracker_for_get_auth, mocker):
    tracker, mocks = make_tracker_for_get_auth()
    tracker._auth = 'my auth'
    auth = await tracker._get_auth()
    assert auth == 'my auth'
    assert mocks.mock_calls == []

@pytest.mark.asyncio
async def test_MtvTracker_get_auth_without_previously_gotten_auth(make_tracker_for_get_auth, mocker):
    tracker, mocks = make_tracker_for_get_auth()
    auth = await tracker._get_auth()
    assert auth is mocks._get_form_value.return_value
    assert mocks.mock_calls == [
        call.login(),
        call.get(tracker._upload_url, user_agent=True),
        call.parse(mocks.get.return_value),
        call._get_form_value(mocks.parse.return_value, 'input', attrs={'name': 'auth'}),
    ]


@pytest.fixture
def make_tracker_for_get_form_value(make_tracker, mocker):
    def make_tracker_for_get_form_value():
        tracker = make_tracker()
        mocks = Mock()
        mocks.attach_mock(mocker.patch.object(tracker, '_raise_error'), '_raise_error')
        return tracker, mocks

    return make_tracker_for_get_form_value

@pytest.mark.parametrize(
    argnames='html, tag_name, kwargs, exp_result',
    argvalues=(
        # Get value from tag: <... value="..."/>
        (
            '<input class="foo" value="asdf"/>',
            'input',
            {'attrs': {'class': 'foo'}},
            'asdf',
        ),

        # Get value from between tags: <textarea>...</textarea>
        (
            '<textarea class="bar">asdf <b>bold</b></textarea>',
            'textarea',
            {'attrs': {'class': 'bar'}},
            'asdf bold',
        ),

        # Empty value: <... value=""/>
        (
            '<input class="foo" value=""/>',
            'input',
            {'attrs': {'class': 'foo'}},
            ValueError('Tag has no value: <input class="foo" value=""/>'),
        ),

        # Empty value: <textarea></textarea>
        (
            '<textarea class="bar"></textarea>',
            'textarea',
            {'attrs': {'class': 'bar'}},
            ValueError('Tag has no value: <textarea class="bar"></textarea>'),
        ),

        # Tag not found
        (
            '<sometag>asdf</sometag>',
            'mytag',
            {},
            ValueError('Could not find tag: mytag'),
        ),

        # Tag class not found
        (
            '<input class="foo" value="asdf"/>',
            'input',
            {'attrs': {'class': 'bar'}},
            ValueError('Could not find tag: input'),
        ),
    ),
)
def test_MtvTracker_get_form_value(html, tag_name, kwargs, exp_result, make_tracker_for_get_form_value):
    tracker, mocks = make_tracker_for_get_form_value()
    doc = utils.html.parse(html)
    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            tracker._get_form_value(doc, tag_name, **kwargs)
    else:
        return_value = tracker._get_form_value(doc, tag_name, **kwargs)
        assert return_value == exp_result


@pytest.fixture
def make_tracker_for_confirm_login(make_tracker, mocker):
    def make_tracker_for_confirm_login():
        tracker = make_tracker()
        mocks = Mock()
        mocks.attach_mock(mocker.patch.object(tracker, '_raise_error'), '_raise_error')
        return tracker, mocks

    return make_tracker_for_confirm_login

@pytest.mark.parametrize(
    argnames='html, exp_user_id, exp_exception',
    argvalues=(
        (
            '<html><a href="/user.php?id=123">superuser</a></html>',
            '123',
            None,
        ),
        (
            '<html><div class="flash error">The <b>ERROR</b> message!</div></html>',
            None,
            errors.RequestError('Login failed: The ERROR message!'),
        ),
        (
            '<html><div>Not the error message</div></html>',
            None,
            RuntimeError('Login failed: No error message found'),
        ),
    ),
)
def test_MtvTracker_confirm_login(html, exp_user_id, exp_exception, make_tracker_for_confirm_login):
    tracker, mocks = make_tracker_for_confirm_login()
    doc = utils.html.parse(html)
    return_value = tracker._confirm_login(doc)
    assert return_value is None
    assert tracker._user_id == exp_user_id

    if exp_exception:
        assert mocks.mock_calls == [
            call._raise_error(doc, msg_prefix='Login failed')
        ]
    else:
        assert mocks.mock_calls == []


@pytest.fixture
def make_tracker_for_raise_error(make_tracker, mocker):
    def make_tracker_for_raise_error():
        tracker = make_tracker()
        mocks = Mock()
        mocks.attach_mock(mocker.patch('upsies.trackers.mtv.tracker.utils.html.dump'), 'dump')
        return tracker, mocks

    return make_tracker_for_raise_error

@pytest.mark.parametrize(
    argnames='html, exp_exception',
    argvalues=(
        (
            '<html><div class="error">Something went wrong</div></html>',
            errors.RequestError('Something went wrong'),
        ),
        (
            '<html><div class="messagebar alert">Something went wrong</div></html>',
            errors.RequestError('Something went wrong'),
        ),
        (
            '<html><div>Something went wrong unexpectedly!</div></html>',
            RuntimeError('No error message found (dumped HTML response to {filepath})'),
        ),
        (
            '',
            errors.RequestError(
                'Zero bytes response bug encountered.\n'
                'You can try again or pray for someone to fix it.\n'
                'Or gain fame and glory by fixing it yourself!'
            ),
        ),
    ),
    ids=lambda v: repr(v),
)
def test_MtvTracker_raise_error(html, exp_exception, make_tracker_for_raise_error):
    tracker, mocks = make_tracker_for_raise_error()
    doc = utils.html.parse(html)
    if exp_exception:
        exp_msg = f'HEADS UP: {exp_exception}'.format(filepath='HEADS UP.html')
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(exp_msg)}$'):
            tracker._raise_error(doc, msg_prefix='HEADS UP')

        if isinstance(exp_exception, RuntimeError):
            assert mocks.mock_calls == [call.dump(doc, 'HEADS UP.html')]
        else:
            assert mocks.mock_calls == []

    else:
        return_value = tracker._raise_error(doc, msg_prefix='HEADS UP')
        assert return_value is None
        assert mocks.mock_calls == []


@pytest.fixture
def make_tracker_for_raise_error_dupes(make_tracker, mocker):
    def make_tracker_for_raise_error_dupes():
        tracker = make_tracker()
        mocks = Mock()
        return tracker, mocks

    return make_tracker_for_raise_error_dupes

@pytest.mark.parametrize(
    argnames='html, exp_exception',
    argvalues=(
        (
            (
                '<div id="messagebar">Hedgehogs found</div>'
                '<table id="not_torrent_table">'
                '<tr class="torrent"><td class="torrent"><a href="torrent.php?123">file1.mkv</a></td></tr>'
                '</table>'
            ),
            None,
        ),
        (
            (
                '<div id="messagebar">Duplicates found</div>'
                '<table id="torrent_table">'
                '</table>'
            ),
            None,
        ),
        (
            (
                '<div id="messagebar">Duplicates found</div>'
                '<table id="not_torrent_table">'
                '<tr class="torrent"><td class="torrent"><a href="torrent.php?123">file1.mkv</a></td></tr>'
                '</table>'
            ),
            None,
        ),
        (
            (
                '<div id="messagebar">Duplicates found</div>'
                '<table id="torrent_table">'
                '<tr class="torrent"><td class="torrent"><a href="torrent.php?123">file1.mkv</a></td></tr>'
                '<tr class="torrent"><td class="torrent"><a href="torrent.php?234">file2.mkv</a></td></tr>'
                '<tr class="torrent"><td class="torrent"><a href="torrent.php?234">file3.mkv</a></td></tr>'
                '</table>'
            ),
            errors.FoundDupeError(['file1.mkv', 'file2.mkv', 'file3.mkv']),
        ),
    ),
)
def test_MtvTracker_raise_error_dupes(html, exp_exception, make_tracker_for_raise_error_dupes):
    tracker, mocks = make_tracker_for_raise_error_dupes()
    doc = utils.html.parse(html)
    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            tracker._raise_error_dupes(doc)
    else:
        return_value = tracker._raise_error_dupes(doc)
        assert return_value is None



import json
import re
import types
from unittest.mock import AsyncMock, Mock, call, patch

import pytest

from upsies import errors, utils
from upsies.utils.predbs import srrdb


@pytest.fixture
def api():
    return srrdb.SrrdbApi()


def test_name():
    assert srrdb.SrrdbApi.name == 'srrdb'


def test_label():
    assert srrdb.SrrdbApi.label == 'srrDB'


def test_default_config():
    assert srrdb.SrrdbApi.default_config == {}


def test__make_release_name_from_path(mocker):
    mocks = Mock()
    mocks.attach_mock(mocker.patch('upsies.utils.fs.basename'), 'basename')
    mocks.attach_mock(mocker.patch('upsies.utils.fs.strip_extension'), 'strip_extension')
    release_name = srrdb.SrrdbApi._make_release_name_from_path('path/to/My.Release.Name')
    assert release_name is mocks.strip_extension.return_value
    assert mocks.mock_calls == [
        call.basename('path/to/My.Release.Name'),
        call.strip_extension(mocks.basename.return_value),
    ]


@pytest.mark.parametrize(
    argnames=(
        'query,'
        'exact_movie_or_episode_results, exact_season_results,'
        'page_size, max_skip, pages, exp_results, exp_mock_calls'
    ),
    argvalues=(
        pytest.param(
            types.SimpleNamespace(release_info='mock release info', keywords='mock keywords', group='mock group'),
            ['The.Exact.Release.Name.720p.BluRay.x264-ASDF'],
            ['The.Exact.Release.Name.S03.720p.BluRay.x264-ASDF'],
            -1, -2, [['pages', 'are'], ['irrelevant', 'here']],
            ['The.Exact.Release.Name.720p.BluRay.x264-ASDF'],
            [
                call._search_for_exact_movie_or_episode(
                    types.SimpleNamespace(release_info='mock release info', keywords='mock keywords', group='mock group')
                ),
            ],
            id='Exact movie or episode name is found',
        ),
        pytest.param(
            types.SimpleNamespace(release_info='mock release info', keywords='mock keywords', group='mock group'),
            [],
            ['The.Exact.Release.Name.S03.720p.BluRay.x264-ASDF'],
            -1, -2, [['pages', 'are'], ['irrelevant', 'here']],
            ['The.Exact.Release.Name.S03.720p.BluRay.x264-ASDF'],
            [
                call._search_for_exact_movie_or_episode(
                    types.SimpleNamespace(release_info='mock release info', keywords='mock keywords', group='mock group')
                ),
                call._search_for_exact_season(
                    types.SimpleNamespace(release_info='mock release info', keywords='mock keywords', group='mock group')
                ),
            ],
            id='Exact season name is found',
        ),
        pytest.param(
            types.SimpleNamespace(release_info=None, keywords='mock keywords', group='mock group'),
            [],
            [],
            3, 99, [['a', 'b', 'c'], []],
            ['a', 'b', 'c'],
            [
                call._get_keywords_path('mock keywords', 'mock group', 0),
                call._request_search_page('keywords/for/page/1'),
                call._get_keywords_path('mock keywords', 'mock group', 3),
                call._request_search_page('keywords/for/page/2'),
            ],
            id='Query does not provide release_info',
        ),
        pytest.param(
            types.SimpleNamespace(release_info='mock release info', keywords='mock keywords', group='mock group'),
            [],
            [],
            3, 99, [['a', 'b', 'c'], []],
            ['a', 'b', 'c'],
            [
                call._search_for_exact_movie_or_episode(
                    types.SimpleNamespace(release_info='mock release info', keywords='mock keywords', group='mock group')
                ),
                call._search_for_exact_season(
                    types.SimpleNamespace(release_info='mock release info', keywords='mock keywords', group='mock group')
                ),
                call._get_keywords_path('mock keywords', 'mock group', 0),
                call._request_search_page('keywords/for/page/1'),
                call._get_keywords_path('mock keywords', 'mock group', 3),
                call._request_search_page('keywords/for/page/2'),
            ],
            id='No exact search results found',
        ),
        pytest.param(
            types.SimpleNamespace(release_info='mock release info', keywords='mock keywords', group='mock group'),
            [],
            [],
            4, 99, [['a', 'b', 'c', 'd'], ['e', 'f', 'g', 'h'], ['i']],
            ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i'],
            [
                call._search_for_exact_movie_or_episode(
                    types.SimpleNamespace(release_info='mock release info', keywords='mock keywords', group='mock group')
                ),
                call._search_for_exact_season(
                    types.SimpleNamespace(release_info='mock release info', keywords='mock keywords', group='mock group')
                ),
                call._get_keywords_path('mock keywords', 'mock group', 0),
                call._request_search_page('keywords/for/page/1'),
                call._get_keywords_path('mock keywords', 'mock group', 4),
                call._request_search_page('keywords/for/page/2'),
                call._get_keywords_path('mock keywords', 'mock group', 8),
                call._request_search_page('keywords/for/page/3'),
            ],
            id='Stop requesting pages if page is not full',
        ),
        pytest.param(
            types.SimpleNamespace(release_info='mock release info', keywords='mock keywords', group='mock group'),
            [],
            [],
            2, 10, [['a', 'b'], ['c', 'd'], ['e', 'f'], ['g', 'h'], ['i', 'j'], ['k', 'l'], ['m', 'n'], ['o', 'p']],
            ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l'],
            [
                call._search_for_exact_movie_or_episode(
                    types.SimpleNamespace(release_info='mock release info', keywords='mock keywords', group='mock group')
                ),
                call._search_for_exact_season(
                    types.SimpleNamespace(release_info='mock release info', keywords='mock keywords', group='mock group')
                ),
                call._get_keywords_path('mock keywords', 'mock group', 0),
                call._request_search_page('keywords/for/page/1'),
                call._get_keywords_path('mock keywords', 'mock group', 2),
                call._request_search_page('keywords/for/page/2'),
                call._get_keywords_path('mock keywords', 'mock group', 4),
                call._request_search_page('keywords/for/page/3'),
                call._get_keywords_path('mock keywords', 'mock group', 6),
                call._request_search_page('keywords/for/page/4'),
                call._get_keywords_path('mock keywords', 'mock group', 8),
                call._request_search_page('keywords/for/page/5'),
                call._get_keywords_path('mock keywords', 'mock group', 10),
                call._request_search_page('keywords/for/page/6'),
            ],
            id='Stop requesting pages if maximum page requests are reached',
        ),
    ),
)
@pytest.mark.asyncio
async def test__search(query,
                       exact_movie_or_episode_results, exact_season_results,
                       page_size, max_skip, pages,
                       exp_results, exp_mock_calls, api, mocker):
    mocker.patch.object(api, '_page_size', page_size)
    mocker.patch.object(api, '_max_skip', max_skip)

    mocks = Mock()
    mocks.attach_mock(
        mocker.patch.object(api, '_search_for_exact_movie_or_episode', return_value=exact_movie_or_episode_results),
        '_search_for_exact_movie_or_episode',
    )
    mocks.attach_mock(
        mocker.patch.object(api, '_search_for_exact_season', return_value=exact_season_results),
        '_search_for_exact_season',
    )
    mocks.attach_mock(
        mocker.patch.object(api, '_get_keywords_path', side_effect=(
            f'keywords/for/page/{i + 1}'
            for i in range(len(pages))
        )),
        '_get_keywords_path',
    )
    mocks.attach_mock(
        mocker.patch.object(api, '_request_search_page', side_effect=pages),
        '_request_search_page',
    )

    response = await api._search(query)
    assert list(response) == exp_results
    assert mocks.mock_calls == exp_mock_calls


@pytest.mark.parametrize(
    argnames='release_type, exp_search_requested',
    argvalues=(
        (utils.release.ReleaseType.movie, True),
        (utils.release.ReleaseType.episode, True),
        (utils.release.ReleaseType.season, False),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.parametrize(
    argnames='release_path, results, exp_results, exp_mock_calls',
    argvalues=(
        pytest.param(
            'path/to/Exact.Release.Name.720p.BluRay.x264-ASDF.mkv',
            ('Exact.Release.Name.720p.BluRay.x264-ASDF',),
            ('Exact.Release.Name.720p.BluRay.x264-ASDF',),
            [
                call._make_release_name_from_path('path/to/Exact.Release.Name.720p.BluRay.x264-ASDF.mkv'),
                call._search_for_exact_release_name('RELEASE_NAME:path/to/Exact.Release.Name.720p.BluRay.x264-ASDF.mkv'),
            ],
            id='Results are found',
        ),
        pytest.param(
            'path/to/Exact.Release.Name.720p.BluRay.x264-ASDF.mkv',
            (),
            (),
            [
                call._make_release_name_from_path('path/to/Exact.Release.Name.720p.BluRay.x264-ASDF.mkv'),
                call._search_for_exact_release_name('RELEASE_NAME:path/to/Exact.Release.Name.720p.BluRay.x264-ASDF.mkv'),
            ],
            id='No results are found',
        ),
    ),
)
@pytest.mark.asyncio
async def test__search_for_exact_movie_or_episode(release_path, release_type, results,
                                                  exp_search_requested, exp_results, exp_mock_calls,
                                                  api, mocker):
    release_name = f'RELEASE_NAME:{release_path}'
    mocks = Mock()
    mocks.attach_mock(
        mocker.patch.object(api, '_make_release_name_from_path', return_value=release_name),
        '_make_release_name_from_path',
    )
    mocks.attach_mock(
        mocker.patch.object(api, '_search_for_exact_release_name', return_value=results),
        '_search_for_exact_release_name',
    )

    query = Mock(release_info=Mock(
        __getitem__=Mock(return_value=release_type),
        path=release_path,
    ))

    results = await api._search_for_exact_movie_or_episode(query)

    if exp_search_requested:
        assert results == exp_results
        assert mocks.mock_calls == exp_mock_calls
    else:
        assert results is None
        assert mocks.mock_calls == []


@pytest.mark.parametrize(
    argnames='is_dir, release_type, exp_search_requested',
    argvalues=(
        pytest.param(False, utils.release.ReleaseType.movie, False, id='nondirectory, movie'),
        pytest.param(False, utils.release.ReleaseType.episode, False, id='nondirectory, episode'),
        pytest.param(False, utils.release.ReleaseType.season, False, id='nondirectory, season'),
        pytest.param(True, utils.release.ReleaseType.movie, False, id='directory, movie'),
        pytest.param(True, utils.release.ReleaseType.episode, False, id='directory, episode'),
        pytest.param(True, utils.release.ReleaseType.season, True, id='directory, season'),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.parametrize(
    argnames='release_path, found_videos, results, exp_results, exp_mock_calls',
    argvalues=(
        pytest.param(
            'path/to/Exact.Release.Name.S03.720p.BluRay.x264-ASDF',
            (
                'path/to/Exact.Release.Name.S03E01.720p.BluRay.x264-ASDF.mkv',
                'path/to/Exact.Release.Name.S03E02.720p.BluRay.x264-ASDF.mkv',
                'path/to/Exact.Release.Name.S03E03.720p.BluRay.x264-ASDF.mkv',
            ),
            {
                'path/to/Exact.Release.Name.S03E01.720p.BluRay.x264-ASDF.mkv': ['Exact.Release.Name.S03E01.720p.BluRay.x264-ASDF'],
                'path/to/Exact.Release.Name.S03E02.720p.BluRay.x264-ASDF.mkv': ['Exact.Release.Name.S03E02.720p.BluRay.x264-ASDF'],
                'path/to/Exact.Release.Name.S03E03.720p.BluRay.x264-ASDF.mkv': ['Exact.Release.Name.S03E03.720p.BluRay.x264-ASDF'],
            },
            [
                'Exact.Release.Name.S03E01.720p.BluRay.x264-ASDF',
                'Exact.Release.Name.S03E02.720p.BluRay.x264-ASDF',
                'Exact.Release.Name.S03E03.720p.BluRay.x264-ASDF',
            ],
            [
                call.find_videos('path/to/Exact.Release.Name.S03.720p.BluRay.x264-ASDF'),
                call._make_release_name_from_path('path/to/Exact.Release.Name.S03E01.720p.BluRay.x264-ASDF.mkv'),
                call._search_for_exact_release_name('RELEASE_NAME:path/to/Exact.Release.Name.S03E01.720p.BluRay.x264-ASDF.mkv'),
                call._make_release_name_from_path('path/to/Exact.Release.Name.S03E02.720p.BluRay.x264-ASDF.mkv'),
                call._search_for_exact_release_name('RELEASE_NAME:path/to/Exact.Release.Name.S03E02.720p.BluRay.x264-ASDF.mkv'),
                call._make_release_name_from_path('path/to/Exact.Release.Name.S03E03.720p.BluRay.x264-ASDF.mkv'),
                call._search_for_exact_release_name('RELEASE_NAME:path/to/Exact.Release.Name.S03E03.720p.BluRay.x264-ASDF.mkv'),
            ],
            id='Results are found',
        ),
        pytest.param(
            'path/to/Exact.Release.Name.S03.720p.BluRay.x264-ASDF',
            (
                'path/to/Exact.Release.Name.S03E01.720p.BluRay.x264-ASDF.mkv',
                'path/to/Exact.Release.Name.S03E02.720p.BluRay.x264-ASDF.mkv',
                'path/to/Exact.Release.Name.S03E03.720p.BluRay.x264-ASDF.mkv',
            ),
            {
                'path/to/Exact.Release.Name.S03E01.720p.BluRay.x264-ASDF.mkv': [],
                'path/to/Exact.Release.Name.S03E02.720p.BluRay.x264-ASDF.mkv': [],
                'path/to/Exact.Release.Name.S03E03.720p.BluRay.x264-ASDF.mkv': [],
            },
            [],
            [
                call.find_videos('path/to/Exact.Release.Name.S03.720p.BluRay.x264-ASDF'),
                call._make_release_name_from_path('path/to/Exact.Release.Name.S03E01.720p.BluRay.x264-ASDF.mkv'),
                call._search_for_exact_release_name('RELEASE_NAME:path/to/Exact.Release.Name.S03E01.720p.BluRay.x264-ASDF.mkv'),
            ],
            id='No results are found',
        ),
        pytest.param(
            'path/to/Exact.Release.Name.S03.720p.BluRay.x264-ASDF',
            (
                'path/to/Exact.Release.Name.S03E01.720p.BluRay.x264-ASDF.mkv',
                'path/to/Exact.Release.Name.S03E02.720p.BluRay.x264-ASDF.mkv',
                'path/to/Exact.Release.Name.S03E03.720p.BluRay.x264-ASDF.mkv',
            ),
            {
                'path/to/Exact.Release.Name.S03E01.720p.BluRay.x264-ASDF.mkv': ['Exact.Release.Name.S03E01.720p.BluRay.x264-ASDF'],
                'path/to/Exact.Release.Name.S03E02.720p.BluRay.x264-ASDF.mkv': [],
                'path/to/Exact.Release.Name.S03E03.720p.BluRay.x264-ASDF.mkv': ['Exact.Release.Name.S03E03.720p.BluRay.x264-ASDF'],
            },
            [],
            [
                call.find_videos('path/to/Exact.Release.Name.S03.720p.BluRay.x264-ASDF'),
                call._make_release_name_from_path('path/to/Exact.Release.Name.S03E01.720p.BluRay.x264-ASDF.mkv'),
                call._search_for_exact_release_name('RELEASE_NAME:path/to/Exact.Release.Name.S03E01.720p.BluRay.x264-ASDF.mkv'),
                call._make_release_name_from_path('path/to/Exact.Release.Name.S03E02.720p.BluRay.x264-ASDF.mkv'),
                call._search_for_exact_release_name('RELEASE_NAME:path/to/Exact.Release.Name.S03E02.720p.BluRay.x264-ASDF.mkv'),
            ],
            id='No results are found for one episode',
        ),
    ),
)
@pytest.mark.asyncio
async def test__search_for_exact_season(release_path, release_type, is_dir, found_videos, results,
                                        exp_search_requested, exp_results, exp_mock_calls,
                                        api, mocker):
    mocks = Mock()
    mocks.attach_mock(
        mocker.patch('upsies.utils.video.find_videos', return_value=found_videos),
        'find_videos',
    )
    mocks.attach_mock(
        mocker.patch.object(api, '_make_release_name_from_path', side_effect=(
            f'RELEASE_NAME:{found_video}'
            for found_video in found_videos
        )),
        '_make_release_name_from_path',
    )
    mocks.attach_mock(
        mocker.patch.object(api, '_search_for_exact_release_name', side_effect=results.values()),
        '_search_for_exact_release_name',
    )

    query = Mock(release_info=Mock(
        __getitem__=Mock(return_value=release_type),
        path=release_path,
    ))

    with patch('os.path.isdir', return_value=is_dir):
        results = await api._search_for_exact_season(query)

    if exp_search_requested:
        assert results == exp_results
        assert mocks.mock_calls == exp_mock_calls
    else:
        assert results is None
        assert mocks.mock_calls == []


@pytest.mark.parametrize(
    argnames='release_name, response, exp_results, exp_mock_calls',
    argvalues=(
        pytest.param(
            'Exact.Release.Name.720p.BluRay.x264-ASDF',
            {'results': [{'release': 'Exact.Release.Name.720p.BluRay.x264-ASDF'}]},
            ('Exact.Release.Name.720p.BluRay.x264-ASDF',),
            [
                call.get(f'{srrdb.SrrdbApi._search_url}/r:Exact.Release.Name.720p.BluRay.x264-ASDF', cache=True),
            ],
            id='Exact release name is found',
        ),
        pytest.param(
            'Exact.Release.Name.720p.BluRay.x264-ASDF',
            {'foo': 'bar'},
            (),
            [
                call.get(f'{srrdb.SrrdbApi._search_url}/r:Exact.Release.Name.720p.BluRay.x264-ASDF', cache=True),
            ],
            id='Nothing is found',
        ),
        pytest.param(
            'Exact.Release.Name.720p.BluRay.x264-ASDF',
            errors.RequestError('What a stupid request.'),
            (),
            [
                call.get(f'{srrdb.SrrdbApi._search_url}/r:Exact.Release.Name.720p.BluRay.x264-ASDF', cache=True),
            ],
            id='Request raises error',
        ),
    ),
)
@pytest.mark.asyncio
async def test__search_for_exact_release_name(release_name, response, exp_results, exp_mock_calls, api, mocker):

    mocks = Mock()
    if isinstance(response, Exception):
        get_mock = AsyncMock(side_effect=response)
    else:
        get_mock = AsyncMock(return_value=Mock(json=Mock(return_value=response)))
    mocks.attach_mock(
        mocker.patch('upsies.utils.http.get', get_mock),
        'get',
    )
    results = await api._search_for_exact_release_name(release_name)
    assert results == exp_results
    assert mocks.mock_calls == exp_mock_calls


@pytest.mark.asyncio
async def test_request_page(api, mocker):
    http_mock = mocker.patch('upsies.utils.http', Mock(
        get=AsyncMock(return_value=Mock(json=Mock(return_value={
            'results': [
                {'release': 'Foo-ASDF'},
                {'release': 'Bar-ASDF'},
                {'release': 'Baz-FDSA'},
            ],
        }))),
    ))

    response = await api._request_search_page('mock/keywords/path')
    assert list(response) == [
        'Foo-ASDF',
        'Bar-ASDF',
        'Baz-FDSA',
    ]

    assert http_mock.get.call_args_list == [
        call(
            f'{api._search_url}/mock/keywords/path',
            cache=True,
        ),
    ]


@pytest.mark.parametrize('skip', (0, 100))
@pytest.mark.parametrize('group', (None, '', 'ASDF'), ids=lambda v: str(v))
@pytest.mark.parametrize(
    argnames='keywords, exp_return_value',
    argvalues=(
        (['10,000', 'BC'], "10/000/bc"),
        (["Lot's", 'of', 'Riff-Raff'], 'lot/s/of/riff/raff'),
    ),
)
def test_get_keywords_path(keywords, group, skip, exp_return_value, api, mocker):
    return_value = api._get_keywords_path(keywords, group=group, skip=skip)

    if group:
        exp_return_value += f'/group:{group.lower()}'
    exp_return_value += '/order:date-desc'
    exp_return_value += f'/skipr:{skip}.{api._page_size}'
    assert return_value == exp_return_value


class MockResponse(str):
    def json(self):
        return json.loads(self)

@pytest.mark.parametrize(
    argnames='release_name, response, exp_result',
    argvalues=(
        pytest.param(
            'Foo.2000-ASDF',
            MockResponse(
                '{'
                '  "name":"Foo.2000-ASDF",'
                '  "archived-files": ['
                '    {"name":"Foo.mkv", "size": 123,"crc": "abc"},'
                '    {"name":"bar.txt", "size": 234,"crc": "def"},'
                '    {"name":"baz.nfo", "size": 456,"crc": "efg"}'
                '  ]'
                '}'
            ),
            {
                'bar.txt': {'release_name': 'Foo.2000-ASDF', 'file_name': 'bar.txt', 'size': 234, 'crc': 'def'},
                'baz.nfo': {'release_name': 'Foo.2000-ASDF', 'file_name': 'baz.nfo', 'size': 456, 'crc': 'efg'},
                'Foo.mkv': {'release_name': 'Foo.2000-ASDF', 'file_name': 'Foo.mkv', 'size': 123, 'crc': 'abc'},
            },
            id='All information is available',
        ),

        pytest.param(
            'Foo.2000-ASDF',
            MockResponse('[]'),
            {},
            id='Response is empty list',
        ),

        pytest.param(
            'Mock.Release.Name',
            MockResponse(''),
            {},
            id='Response is empty string',
        ),
    ),
)
@pytest.mark.asyncio
async def test_release_files(release_name, response, exp_result, api, mocker):
    http_get_mock = mocker.patch('upsies.utils.http.get', return_value=response)

    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            await api._release_files(release_name)
    else:
        return_value = await api._release_files(release_name)
        assert return_value == exp_result

    assert http_get_mock.call_args_list == [call(
        f'{api._details_url}/{release_name}',
        cache=True,
    )]



"""
Commands related to torrent files
"""

import functools
import os

from .... import constants, jobs, trackers, utils
from .base import CommandBase


class torrent_add(CommandBase):
    """Add torrent file to BitTorrent client"""

    names = ('torrent-add', 'ta')

    argument_definitions = {
        'CLIENT': {
            'type': utils.argtypes.client,
            'help': ('Case-insensitive BitTorrent client name\n'
                     'Supported clients: ' + ', '.join(utils.btclient.client_names())),
        },
        'TORRENT': {
            'nargs': '+',
            'help': 'Path to torrent file',
        },
        ('--download-path', '-p'): {
            'help': "Parent directory of the torrent's content",
            'metavar': 'DIRECTORY',
        },
    }

    @functools.cached_property
    def jobs(self):
        client_name = self.args.CLIENT
        options = self.get_options('clients', client_name)
        return (
            jobs.torrent.AddTorrentJob(
                home_directory=self.home_directory,
                cache_directory=self.cache_directory,
                ignore_cache=self.args.ignore_cache,
                btclient=utils.btclient.BtClient(
                    name=client_name,
                    url=options['url'],
                    username=options['username'],
                    password=options['password'],
                    download_path=self.args.download_path,
                    check_after_add=options['check_after_add'],
                    category=options.get('category', ''),
                ),
                enqueue=self.args.TORRENT,
            ),
        )


class torrent_create(CommandBase):
    """
    Create torrent file and optionally add or copy it

    The piece hashes are cached and re-used if possible. This means the torrent
    is not generated for every tracker again. Note that the torrent is always
    generated if files are excluded on one tracker but not on the other. The
    files in each torrent must be identical.
    """

    names = ('torrent-create', 'tc')

    argument_definitions = {}

    subcommand_name = 'TRACKER'
    subcommands = {
        tracker.name: {
            'description': (
                f'Create a torrent file for {tracker.label} '
                'in the current working directory.'
            ),
            'cli': {
                # Default arguments for all tackers
                **{
                    'CONTENT': {
                        'type': utils.argtypes.content,
                        'help': 'Path to release content',
                    },
                    ('--exclude-files', '--ef'): {
                        'nargs': '+',
                        'metavar': 'PATTERN',
                        'help': ('Glob pattern to exclude from torrent '
                                 '(matched case-insensitively against path in torrent)'),
                        'default': (),
                    },
                    ('--exclude-files-regex', '--efr'): {
                        'nargs': '+',
                        'metavar': 'PATTERN',
                        'help': ('Regular expression to exclude from torrent '
                                 '(matched case-sensitively against path in torrent)'),
                        'type': utils.argtypes.regex,
                        'default': (),
                    },
                    ('--reuse-torrent', '-t'): {
                        'nargs': '+',
                        'metavar': 'TORRENT',
                        'help': ('Use hashed pieces from TORRENT instead of generating '
                                 'them again or getting them from '
                                 f'{utils.fs.tildify_path(constants.GENERIC_TORRENTS_DIRPATH)}\n'
                                 'TORRENT may also be a directory, which is searched recursively '
                                 'for a matching *.torrent file.\n'
                                 "NOTE: This option is ignored if TORRENT doesn't match CONTENT properly."),
                        'type': utils.argtypes.existing_path,
                        'default': (),
                    },
                    ('--add-to', '-a'): {
                        'type': utils.argtypes.client,
                        'metavar': 'CLIENT',
                        'help': ('Case-insensitive BitTorrent client name\n'
                                 'Supported clients: ' + ', '.join(utils.btclient.client_names())),
                    },
                    ('--copy-to', '-c'): {
                        'metavar': 'PATH',
                        'help': 'Copy the created torrent to PATH (file or directory)',
                    },
                },
                # Custom arguments defined by tracker for this command
                **tracker.TrackerConfig.argument_definitions.get('torrent-create', {}),
            }
        }
        for tracker in trackers.trackers()
    }

    @functools.cached_property
    def tracker_name(self):
        return self.args.subcommand.lower()

    @property
    def home_directory(self):
        """Create torrent file in current working directory"""
        return '.'

    @functools.cached_property
    def create_torrent_job(self):
        return jobs.torrent.CreateTorrentJob(
            home_directory=self.home_directory,
            cache_directory=self.cache_directory,
            ignore_cache=self.args.ignore_cache,
            content_path=self.args.CONTENT,
            reuse_torrent_path=(
                tuple(self.args.reuse_torrent)
                + tuple(self.config['config']['torrent-create']['reuse_torrent_paths'])
            ),
            tracker=trackers.tracker(
                name=self.tracker_name,
                options={**self.config['trackers'][self.tracker_name],
                         **vars(self.args)},
            ),
            exclude_files=(
                tuple(self.args.exclude_files)
                + tuple(self.args.exclude_files_regex)
            ),
        )

    @functools.cached_property
    def add_torrent_job(self):
        if self.args.add_to:
            client_name = self.args.add_to
            options = self.get_options('clients', client_name)
            add_torrent_job = jobs.torrent.AddTorrentJob(
                home_directory=self.home_directory,
                cache_directory=self.cache_directory,
                ignore_cache=self.args.ignore_cache,
                btclient=utils.btclient.BtClient(
                    name=client_name,
                    url=options['url'],
                    username=options['username'],
                    password=options['password'],
                    download_path=utils.fs.dirname(os.path.abspath(self.args.CONTENT)),
                    check_after_add=options['check_after_add'],
                    category=options.get('category', ''),
                ),
            )

            # Pass CreateTorrentJob output to AddTorrentJob input.
            self.create_torrent_job.signal.register('output', add_torrent_job.enqueue)

            # Finish AddTorrentJob when CreateTorrentJob is finished. The
            # torrent will be enqueued and AddTorrentJob will finish after
            # adding the enqueued torrent.
            self.create_torrent_job.signal.register('finished', lambda _: add_torrent_job.finalize())

            return add_torrent_job

    @functools.cached_property
    def copy_torrent_job(self):
        if self.args.copy_to:
            copy_torrent_job = jobs.torrent.CopyTorrentJob(
                home_directory=self.home_directory,
                cache_directory=self.cache_directory,
                ignore_cache=self.args.ignore_cache,
                destination=self.args.copy_to,
            )
            # Pass CreateTorrentJob output to CopyTorrentJob input.
            self.create_torrent_job.signal.register('output', copy_torrent_job.enqueue)
            # Tell CopyTorrentJob to finish when CreateTorrentJob is done.
            self.create_torrent_job.signal.register('finished', lambda _: copy_torrent_job.finalize())
            return copy_torrent_job

    @functools.cached_property
    def jobs(self):
        return (
            self.create_torrent_job,
            self.add_torrent_job,
            self.copy_torrent_job,
        )



import base64

import pytest

from upsies import utils
from upsies.trackers.bhd import BhdTrackerConfig


@pytest.fixture
def tracker_config():
    return BhdTrackerConfig()


def test_BhdTrackerConfig_defaults(tracker_config):
    assert set(tracker_config) == {
        'upload_url',
        'announce_url',
        'announce_passkey',
        'apikey',
        'source',
        'randomize_infohash',
        'anonymous',
        'draft',
        'image_host',
        'screenshots',
        'exclude',

        # Inherited from TrackerConfigBase
        'add_to',
        'copy_to',
    }


def test_BhdTrackerConfig_defaults_upload_url(tracker_config):
    assert tracker_config['upload_url'] == base64.b64decode('aHR0cHM6Ly9iZXlvbmQtaGQubWUvYXBpL3VwbG9hZA==').decode('ascii')


def test_BhdTrackerConfig_defaults_announce_url(tracker_config):
    assert tracker_config['announce_url'] == base64.b64decode('aHR0cHM6Ly9iZXlvbmQtaGQubWUvYW5ub3VuY2U=').decode('ascii')
    assert tracker_config['announce_url'].description == (
        'The announce URL without the private passkey.'
    )


def test_BhdTrackerConfig_defaults_announce_passkey(tracker_config):
    assert tracker_config['announce_passkey'] == ''
    assert tracker_config['announce_passkey'].description == (
        'The private part of the announce URL.\n'
        'Get it from the website: My Security -> Passkey'
    )


def test_BhdTrackerConfig_defaults_apikey(tracker_config):
    assert tracker_config['apikey'] == ''
    assert tracker_config['apikey'].description == (
        'Your personal private API key.\n'
        'Get it from the website: My Security -> API key'
    )


def test_BhdTrackerConfig_defaults_source(tracker_config):
    assert tracker_config['source'] == 'BHD'


def test_BhdTrackerConfig_defaults_anonymous(tracker_config):
    assert isinstance(tracker_config['anonymous'], utils.types.Bool)
    assert not tracker_config['anonymous']
    assert tracker_config['anonymous'] == 'no'
    assert tracker_config['anonymous'].description == (
        'Whether your username is displayed on your uploads.'
    )


def test_BhdTrackerConfig_defaults_draft(tracker_config):
    assert isinstance(tracker_config['draft'], utils.types.Bool)
    assert not tracker_config['draft']
    assert tracker_config['draft'] == 'no'


def test_BhdTrackerConfig_defaults_image_host(tracker_config, assert_config_list_of_choice):
    exp_options = ('imgbox', 'ptpimg', 'imgbb')
    assert_config_list_of_choice(
        items=tracker_config['image_host'],
        exp_items=('imgbox',),
        exp_options=exp_options,
        exp_description=(
            'List of image hosting service names. The first service is normally used '
            + 'with the others as backup if uploading to the first fails.\n'
            + 'Supported services: ' + ', '.join(sorted(exp_options))
        ),
    )


def test_BhdTrackerConfig_defaults_screenshots(tracker_config, assert_config_number):
    assert_config_number(
        number=tracker_config['screenshots'],
        value=4,
        min=3,
        max=10,
        description='How many screenshots to make.',
    )


def test_BhdTrackerConfig_defaults_exclude(tracker_config):
    assert tracker_config['exclude'] == (
        utils.types.RegEx(r'\.(?i:nfo|txt|jpg|jpeg|png|sfv|md5|idx|sub|srt)$'),
        utils.types.RegEx(r'/(?i:sample|extra|bonus|feature)'),
        utils.types.RegEx(r'(?i:.*sample\.[a-z]+)$'),
    )


def test_BhdTrackerConfig_arguments(tracker_config):
    exp_argument_definitions = {
        'submit': {
            ('--custom-edition', '--ce'),
            ('--draft', '--dr'),
            ('--personal-rip', '--pr'),
            ('--screenshots', '--ss'),
            ('--special', '--sp'),
            ('--only-description', '--od'),
            ('--only-title', '--ot'),
        },
    }
    assert set(tracker_config.argument_definitions) == set(exp_argument_definitions)
    for command in exp_argument_definitions:
        assert set(tracker_config.argument_definitions[command]) == exp_argument_definitions[command]


def test_BhdTrackerConfig_argument_definitions_submit_custom_edition(tracker_config):
    assert tracker_config.argument_definitions['submit'][('--custom-edition', '--ce')] == {
        'help': 'Non-standard edition, e.g. "Final Cut"',
        'default': '',
    }


def test_BhdTrackerConfig_argument_definitions_submit_draft(tracker_config):
    assert tracker_config.argument_definitions['submit'][('--draft', '--dr')] == {
        'help': 'Upload as draft',
        'action': 'store_true',
        'default': None,
    }


def test_BhdTrackerConfig_argument_definitions_submit_personal_rip(tracker_config):
    assert tracker_config.argument_definitions['submit'][('--personal-rip', '--pr')] == {
        'help': 'Tag as your own encode',
        'action': 'store_true',
    }


def test_BhdTrackerConfig_argument_definitions_submit_screenshots(tracker_config):
    assert tracker_config.argument_definitions['submit'][('--screenshots', '--ss')] == {
        'help': ('How many screenshots to make '
                 f'(min={tracker_config["screenshots"].min}, '
                 f'max={tracker_config["screenshots"].max})'),
        'type': utils.argtypes.number_of_screenshots(
            min=tracker_config['screenshots'].min,
            max=tracker_config['screenshots'].max,
        ),
    }


def test_BhdTrackerConfig_argument_definitions_submit_special(tracker_config):
    assert tracker_config.argument_definitions['submit'][('--special', '--sp')] == {
        'help': 'Tag as special episode, e.g. Christmas special (ignored for movie uploads)',
        'action': 'store_true',
    }


def test_BhdTrackerConfig_argument_definitions_submit_only_description(tracker_config):
    assert tracker_config.argument_definitions['submit'][('--only-description', '--od')] == {
        'help': 'Only generate description (do not upload anything)',
        'action': 'store_true',
        'group': 'generate-metadata',
    }


def test_BhdTrackerConfig_argument_definitions_submit_only_title(tracker_config):
    assert tracker_config.argument_definitions['submit'][('--only-title', '--ot')] == {
        'help': 'Only generate title (do not upload anything)',
        'action': 'store_true',
        'group': 'generate-metadata',
    }



import re
import types
from unittest.mock import AsyncMock, Mock, call

import pytest

from upsies import constants, errors, utils
from upsies.utils.predbs import SceneQuery, base


def make_TestDb(default_config=None, **kwargs):
    default_config_ = default_config

    class TestDb(base.PredbApiBase):
        name = 'scn'
        label = 'SCN'
        default_config = default_config_ or {}
        _search = AsyncMock()
        _release_files = AsyncMock()

    return TestDb(**kwargs)


def test_config_property():
    db = make_TestDb()
    assert db.config == {}
    db = make_TestDb(default_config={'foo': 1, 'bar': 2})
    assert db.config == {'foo': 1, 'bar': 2}
    db = make_TestDb(default_config={'foo': 1, 'bar': 2}, config={'bar': 99})
    assert db.config == {'foo': 1, 'bar': 99}


@pytest.mark.parametrize(
    argnames='query, only_existing_releases, results, exp_mock_calls, exp_return_value_from',
    argvalues=(
        pytest.param(
            'path/to/My.Release.2000.Bluray, x264-ASDF',
            'mock_only_existing_releases',
            ['Foo', 'Bar', 'Baz'],
            [
                call._search(SceneQuery('My', 'Release', '2000', 'BluRay', 'x264', group='ASDF')),
                call._postprocess_search_results(
                    ['Foo', 'Bar', 'Baz'],
                    SceneQuery('My', 'Release', '2000', 'BluRay', 'x264', group='ASDF'),
                    'mock_only_existing_releases',
                ),
            ],
            '_postprocess_search_results',
            id='[non-season-pack] Query is path',
        ),
        pytest.param(
            {
                'type': utils.release.ReleaseType.movie,
                'title': 'My Release',
                'source': 'BluRay',
                'group': 'ASDF',
            },
            'mock_only_existing_releases',
            ['Foo', 'Bar', 'Baz'],
            [
                call._search(SceneQuery('My', 'Release', 'BluRay', group='ASDF')),
                call._postprocess_search_results(
                    ['Foo', 'Bar', 'Baz'],
                    SceneQuery('My', 'Release', 'BluRay', group='ASDF'),
                    'mock_only_existing_releases',
                ),
            ],
            '_postprocess_search_results',
            id='[non-season-pack] Query is mapping',
        ),
        pytest.param(
            utils.predbs.SceneQuery('My', 'Release', 'BluRay', group='ASDF'),
            'mock_only_existing_releases',
            ['Foo', 'Bar', 'Baz'],
            [
                call._search(SceneQuery('My', 'Release', 'BluRay', group='ASDF')),
                call._postprocess_search_results(
                    ['Foo', 'Bar', 'Baz'],
                    SceneQuery('My', 'Release', 'BluRay', group='ASDF'),
                    'mock_only_existing_releases',
                ),
            ],
            '_postprocess_search_results',
            id='[non-season-pack] Query is SceneQuery',
        ),
        pytest.param(
            'path/to/My.Release.S03.Bluray.x264-ASDF',
            'mock_only_existing_releases',
            [],
            [
                call._search(SceneQuery('My', 'Release', 'BluRay', 'x264', group='ASDF', episodes={'3': []})),
                call._search_for_episodes(
                    'path/to/My.Release.S03.Bluray.x264-ASDF',
                    'mock_only_existing_releases',
                ),
            ],
            '_search_for_episodes',
            id='[season-pack] Query is path',
        ),
        pytest.param(
            {
                'type': utils.release.ReleaseType.movie,
                'title': 'My Release',
                'episodes': {'3': []},
                'source': 'BluRay',
                'group': 'ASDF',
            },
            'mock_only_existing_releases',
            [],
            [
                call._search(SceneQuery('My', 'Release', 'BluRay', group='ASDF', episodes={'3': []})),
                call._postprocess_search_results(
                    [],
                    SceneQuery('My', 'Release', 'BluRay', group='ASDF', episodes={'3': []}),
                    'mock_only_existing_releases',
                ),
            ],
            '_postprocess_search_results',
            id='[season-pack] Query is mapping',
        ),
    ),
)
@pytest.mark.asyncio
async def test_search(query, results, only_existing_releases, exp_mock_calls, exp_return_value_from, mocker):
    db = make_TestDb()
    mocks = Mock()
    mocks.attach_mock(
        mocker.patch.object(db, '_search', AsyncMock(return_value=results)),
        '_search',
    )
    mocks.attach_mock(
        mocker.patch.object(db, '_postprocess_search_results'),
        '_postprocess_search_results',
    )
    mocks.attach_mock(
        mocker.patch.object(db, '_search_for_episodes'),
        '_search_for_episodes',
    )

    return_value = await db.search(
        query=query,
        only_existing_releases=only_existing_releases,
    )
    exp_return_value = getattr(mocks, exp_return_value_from).return_value
    assert return_value is exp_return_value
    assert mocks.mock_calls == exp_mock_calls

@pytest.mark.asyncio
async def test_search_is_not_supported(mocker):
    db = make_TestDb()
    db._search.side_effect = NotImplementedError()
    with pytest.raises(errors.RequestError, match=fr'{re.escape(str(db.name))} does not support searching'):
        await db.search('foo')


@pytest.mark.asyncio
async def test_search_for_episodes(mocker):
    db = make_TestDb()
    mocks = Mock()
    mocks.attach_mock(
        mocker.patch.object(db, '_generate_episode_queries', Mock(return_value=iter((
            'mock episode_query 1',
            'mock episode_query 2',
            'mock episode_query 3',
        )))),
        '_generate_episode_queries',
    )
    mocks.attach_mock(
        mocker.patch.object(db, 'search', side_effect=(
            ['results 1 for episode_query 1', 'results 2 for episode_query 1', 'results 1 for episode_query 1'],
            [],
            ['results 1 for episode_query 3', 'results 2 for episode_query 3'],
        )),
        'search',
    )
    mock_only_existing_releases = 'mock only_existing_releases'
    path = 'path/to/season.pack'

    return_value = await db._search_for_episodes(path, mock_only_existing_releases)
    assert return_value == [
        'results 1 for episode_query 1', 'results 2 for episode_query 1', 'results 1 for episode_query 1',
        'results 1 for episode_query 3', 'results 2 for episode_query 3',
    ]
    assert mocks.mock_calls == [
        call._generate_episode_queries(path),
        call.search('mock episode_query 1', mock_only_existing_releases),
        call.search('mock episode_query 2', mock_only_existing_releases),
        call.search('mock episode_query 3', mock_only_existing_releases),
    ]

@pytest.mark.parametrize(
    argnames='path, exp_queries_generated',
    argvalues=(
        ('path/to/Foo.2000.x264-ASDF', False),
        ('path/to/Foo.S03E06.x264-ASDF.mkv', False),
        ('path/to/Foo.S03.x264-MiXED/', True),
    ),
)
def test_generate_episode_queries(path, exp_queries_generated, mocker):
    db = make_TestDb()

    file_list = [
        'Foo.S01E01.x264-ASDF.mkv',
        'arf-foo-s01e02.mkv',
        'Foo.S01E03.x264-FDSA.mkv',
    ]

    mocks = Mock()
    mocks.attach_mock(
        mocker.patch('upsies.utils.fs.file_list', Mock(return_value=file_list)),
        'file_list',
    )

    return_value = db._generate_episode_queries(path)
    assert isinstance(return_value, types.GeneratorType)

    if not exp_queries_generated:
        assert list(return_value) == []
    else:
        assert list(return_value) == [
            utils.predbs.SceneQuery.from_string(file_list[0]),
            utils.predbs.SceneQuery.from_string(file_list[2]),
        ]
        assert mocks.mock_calls == [
            call.file_list(path, extensions=constants.VIDEO_FILE_EXTENSIONS),
        ]


@pytest.mark.asyncio
async def test_release_files_is_not_supported(mocker):
    db = make_TestDb()
    db._release_files.side_effect = NotImplementedError()
    with pytest.raises(errors.RequestError, match=fr'{re.escape(str(db.name))} does not provide file information'):
        await db.release_files('foo')



"""
API for trackers

:class:`~.base.TrackerBase` provides a uniform interface for all trackers. Its
subclasses have two main purposes:

1. Specify jobs that generate metadata, e.g. torrent creation, ``mediainfo``
   output, IMDb ID, etc.

2. Provide coroutine methods, e.g. for uploading the generated metadata.
"""

from .. import utils
from . import bhd, dummy, mtv, nbl, ptp
from .base import TrackerBase, TrackerConfigBase, TrackerJobsBase


def trackers():
    """Return list of :class:`.TrackerBase` subclasses"""
    return utils.subclasses(TrackerBase, utils.submodules(__package__))


def tracker(name, **kwargs):
    """
    Create :class:`.TrackerBase` instance

    :param str name: Name of the tracker. A subclass of :class:`.TrackerBase`
        with the same :attr:`~.TrackerBase.name` must exist in one of this
        package's submodules.
    :param kwargs: All keyword arguments are passed to the subclass specified by
        `name`

    :raise ValueError: if no matching subclass can be found

    :return: :class:`.TrackerBase` instance
    """
    for cls in trackers():
        if cls.name == name:
            return cls(**kwargs)
    raise ValueError(f'Unsupported tracker: {name}')


def tracker_names():
    """Return sequence of valid `name` arguments for :func:`.tracker`"""
    return sorted(utils.CaseInsensitiveString(cls.name) for cls in trackers())



from base64 import b64decode

from ... import errors
from .. import http
from . import base

import logging  # isort:skip
_log = logging.getLogger(__name__)


class PredbdeApi(base.PredbApiBase):
    name = 'predbde'
    label = 'PreDB.de'

    default_config = {}

    _url_base = b64decode('cHJlZGIuZGU=').decode('ascii')
    _search_url = f'https://{_url_base}/api/'

    async def _search(self, query):
        q = self._get_q(query.keywords, query.group)
        if q:
            return await self._request_all_pages(q)
        else:
            return ()

    def _get_q(self, keywords, group):
        kws = list(keywords)
        if group:
            kws.append(group)
        if kws:
            kws = (str(kw).lower().strip() for kw in kws)
            return ' '.join(kw for kw in kws if kw)
        else:
            return ''

    _max_pages = 1000

    async def _request_all_pages(self, q):
        combined_results = []
        page = 1
        while page < self._max_pages:
            results, next_page = await self._request_page(q, page)
            combined_results.extend(results)

            if next_page < 0:
                # Negative next page means there are no more pages
                break
            else:
                page += 1

        return combined_results

    _max_results_per_page = 20

    async def _request_page(self, q, page):

        params = {
            'q': q,
            'page': page,
        }
        _log.debug('%s search: %r, %r', self.label, self._search_url, params)
        response = (await http.get(self._search_url, params=params, cache=True)).json()

        if response.get('data'):
            # We found at least one release
            results = tuple(result['release'] for result in response['data'])

        elif (
                response.get('status') != 'success'
                and response.get('message')
                and response['message'].lower() != 'no results'
        ):
            # Report error from API
            raise errors.RequestError(f'{self.label}: {response["message"]}')

        else:
            results = ()

        # Is there another page of results?
        if len(results) >= self._max_results_per_page:
            next_page = page + 1
        else:
            next_page = -1

        return results, next_page

    async def _release_files(self, release_name):
        raise NotImplementedError()



import pytest

from upsies.trackers.ptp import metadata


@pytest.mark.parametrize(
    argnames='reason, exp_value',
    argvalues=(
        (metadata.PtpTrumpableReason.NO_ENGLISH_SUBTITLES, 14),
        (metadata.PtpTrumpableReason.HARDCODED_SUBTITLES, 4),
    ),
    ids=lambda v: repr(v),
)
def test_PtpTrumpableReason_value(reason, exp_value):
    assert reason.value == exp_value


@pytest.mark.parametrize(
    argnames='reason, string',
    argvalues=(
        (metadata.PtpTrumpableReason.NO_ENGLISH_SUBTITLES, 'No English Subtitles'),
        (metadata.PtpTrumpableReason.HARDCODED_SUBTITLES, 'Hardcoded Subtitles'),
    ),
    ids=lambda v: repr(v),
)
def test_PtpTrumpableReason_string(reason, string):
    assert str(reason) == string
    assert type(reason).from_string(string) is reason



"""
Swiss Army knife
"""

import asyncio
import collections
import functools
import hashlib
import importlib
import inspect
import itertools
import os
import re
import types as _types


def get_aioloop():
    """Return :class:`asyncio.AbstractEventLoop` instance"""
    # https://docs.python.org/3.10/library/asyncio-eventloop.html
    try:
        return asyncio.get_running_loop()
    # "no running event loop"
    except RuntimeError:
        # We need a loop before the application has started.
        # This mimics get_event_loop(), but that is going to be an alias for
        # get_running_loop() in future Python releases.
        return asyncio.get_event_loop_policy().get_event_loop()


def os_family():
    """
    Return "windows" or "unix"
    """
    return 'windows' if os.name == 'nt' else 'unix'


# https://github.com/tensorflow/tensorflow/blob/master/tensorflow/python/util/lazy_loader.py
class LazyModule(_types.ModuleType):
    """
    Lazily import module to decrease execution time

    :param str module: Name of the module
    :param mapping namespace: Usually the return value of `globals()`
    :param str name: Name of the module in `namespace`; defaults to `module`
    """

    def __init__(self, module, namespace, name=None):
        self._module = module
        self._namespace = namespace
        self._name = name or module
        super().__init__(module)

    def _load(self):
        # Import the target module and insert it into the parent's namespace
        module = importlib.import_module(self.__name__)
        self._namespace[self._name] = module

        # Update this object's dict so that if someone keeps a reference to the
        # LazyLoader, lookups are efficient (__getattr__ is only called on
        # lookups that fail).
        self.__dict__.update(module.__dict__)

        return module

    def __getattr__(self, item):
        module = self._load()
        return getattr(module, item)

    def __dir__(self):
        module = self._load()
        return dir(module)


def submodules(package):
    """
    Return list of submodules and subpackages in `package`

    :param str package: Fully qualified name of parent package,
        e.g. "upsies.utils.imghosts"
    """
    # Get absolute path to parent directory of top-level package
    own_path = os.path.dirname(__file__)
    rel_path = __package__.replace('.', os.sep)
    assert own_path.endswith(rel_path), f'{own_path!r}.endswith({rel_path!r})'
    project_path = own_path[:-len(rel_path)]

    # Add relative path within project to given package
    package_path = os.path.join(project_path, package.replace('.', os.sep))

    # Find and import public submodules
    submods = []
    for name in os.listdir(package_path):
        if not name.startswith('_'):
            if name.endswith('.py'):
                name = name[:-3]
            if '.' not in name:
                submods.append(
                    importlib.import_module(name=f'.{name}', package=package)
                )
    return submods


def subclasses(basecls, modules):
    """
    Find subclasses in modules

    :param type basecls: Class that all returned classes are a subclass of
    :param modules: Modules to search
    :type modules: list of module objects
    """
    subclses = set()
    for mod in modules:
        for name, member in inspect.getmembers(mod):
            if (member is not basecls and
                isinstance(member, type) and
                issubclass(member, basecls)):
                subclses.add(member)
    return subclses


def closest_number(n, ns, max=None, default=0):
    """
    Return the number from `ns` that is closest to `n`

    :param n: Given number
    :param ns: Sequence of allowed numbers
    :param max: Remove any item from `ns` that is larger than `max`
    :param default: Return value in case `ns` is empty
    """
    if max is not None:
        ns_ = tuple(n_ for n_ in ns if n_ <= max)
        if not ns_:
            raise ValueError(f'No number equal to or below {max}: {ns}')
    else:
        ns_ = ns
    return min(ns_, key=lambda x: abs(x - n), default=default)


class CaseInsensitiveString(str):
    """String that ignores case when compared or sorted"""

    def __hash__(self):
        return hash(self.casefold())

    def __eq__(self, other):
        if not isinstance(other, str):
            return NotImplemented
        else:
            return self.casefold() == other.casefold()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        return self.casefold() < other.casefold()

    def __le__(self, other):
        return self.casefold() <= other.casefold()

    def __gt__(self, other):
        return self.casefold() > other.casefold()

    def __ge__(self, other):
        return self.casefold() >= other.casefold()


class MonitoredList(collections.abc.MutableSequence):
    """
    :class:`list` that calls `callback` after every change

    :param callback: Callable that gets the instance as a positional argument
    """

    def __init__(self, *args, callback, **kwargs):
        self._list = list(*args, **kwargs)
        self._callback = callback

    def __getitem__(self, index):
        return self._list[index]

    def __setitem__(self, index, value):
        self._list[index] = value
        self._callback(self)

    def __delitem__(self, index):
        del self._list[index]
        self._callback(self)

    def insert(self, index, value):
        self._list.insert(index, value)
        self._callback(self)

    def __len__(self):
        return len(self._list)

    def __eq__(self, other):
        return self._list == other

    def __repr__(self):
        return f'{type(self).__name__}({repr(self._list)}, callback={self._callback!r})'


class ImmutableMapping(collections.abc.Mapping):
    """
    :class:`dict`-like that can't be changed after initialization

    All values are also made immutable recursively. Keys should already be
    immutable because the must be hashable.
    """

    def __init__(self, *args, **kwargs):
        self._dct = {
            k: self._make_immutable(v)
            for k, v in dict(*args, **kwargs).items()
        }

    def _make_immutable(self, obj):
        if isinstance(obj, str):
            # Catch strings early because they are also sequences
            return obj

        elif isinstance(obj, collections.abc.Sequence):
            return tuple(self._make_immutable(item) for item in obj)

        elif isinstance(obj, collections.abc.Mapping):
            return ImmutableMapping(obj)

        else:
            return obj

    def __getitem__(self, key):
        return self._dct[key]

    def __iter__(self):
        return iter(self._dct)

    def __len__(self):
        return len(self._dct)

    def __hash__(self):
        return hash(tuple(frozenset(self._dct.items())))

    def __eq__(self, other):
        if isinstance(other, collections.abc.Mapping):
            return hash(self) == hash(tuple(frozenset(self._dct.items())))
        else:
            return NotImplemented

    def __repr__(self):
        return f'{type(self).__name__}({repr(self._dct)})'


def is_sequence(obj):
    """Return whether `obj` is a sequence and not a string"""
    return (
        isinstance(obj, collections.abc.Sequence)
        and not isinstance(obj, str)
    )


def merge_dicts(a, b, path=()):
    """
    Merge nested dictionaries `a` and `b` into new dictionary with same
    structure
    """
    keys = itertools.chain(a, b)
    merged = {}
    for key in keys:
        if (isinstance(a.get(key), collections.abc.Mapping) and
            isinstance(b.get(key), collections.abc.Mapping)):
            merged[key] = merge_dicts(a[key], b[key], path + (key,))
        elif key in b:
            # Value from b takes precedence
            merged[key] = b[key]
        elif key in a:
            # Value from a is default
            merged[key] = a[key]
    return merged


def deduplicate(seq, key=None):
    """
    Return sequence `seq` with all duplicate items removed while maintaining
    the original order

    :param key: Callable that gets each item and returns a hashable identifier
        for that item
    """
    if key is None:
        def key(k):
            return k

    seen_keys = set()
    deduped = []
    for item in seq:
        k = key(item)
        if k not in seen_keys:
            seen_keys.add(k)
            deduped.append(item)
    return deduped


def as_groups(sequence, group_sizes, default=None):
    """
    Iterate over items from `sequence` in equally sized groups

    :params sequence: List of items to group
    :params group_sizes: Sequence of acceptable items in a group

        Determine the group size with the lowest number of `default` items in
        the last group. That group size is then used for all groups.
    :param default: Value to pad last group with if
        ``len(sequence) % group_size != 0``

    Example:

    >>> sequence = range(1, 10)
    >>> for group in as_groups(sequence, [4, 5], default='_'):
    ...     print(group)
    (1, 2, 3, 4, 5)
    (6, 7, 8, 9, '_')
    >>> for group in as_groups(sequence, [3, 4], default='_'):
    ...     print(group)
    (1, 2, 3)
    (4, 5, 6)
    (7, 8, 9)
    """
    # Calculate group size that results in the least number of `default` values
    # in the final group
    gs_map = collections.defaultdict(lambda: [])
    for gs in group_sizes:
        # How many items from `sequence` are in the last group
        overhang = len(sequence) % gs
        # How many `default` values are in the last group
        default_count = 0 if overhang == 0 else gs - overhang
        gs_map[default_count].append(gs)

    lowest_default_count = sorted(gs_map)[0]
    group_size = max(gs_map[lowest_default_count])
    args = [iter(sequence)] * group_size
    yield from itertools.zip_longest(*args, fillvalue=default)


def is_regex_pattern(object):
    """
    Whether `object` is a regular expression object

    A regular expression object is usually obtained from :func:`re.compile`.
    """
    return isinstance(object, re.Pattern)


_unsupported_semantic_hash_types = (
    collections.abc.Iterator,
    collections.abc.Iterable,
    collections.abc.Generator,
)

def semantic_hash(obj):
    """
    Return SHA256 hash for `obj` that stays the same between Python interpreter
    sessions

    https://github.com/schollii/sandals/blob/master/json_sem_hash.py
    """
    def as_str(obj):
        if isinstance(obj, str):
            return obj

        elif isinstance(obj, collections.abc.Mapping):
            stringified = ((as_str(k), as_str(v)) for k, v in obj.items())
            return as_str(sorted(stringified))

        elif isinstance(obj, (collections.abc.Sequence, collections.abc.Set)):
            stringified = (as_str(item) for item in obj)
            return ''.join(sorted(stringified))

        elif isinstance(obj, _unsupported_semantic_hash_types):
            raise RuntimeError(f'Unsupported type: {type(obj)}: {obj!r}')

        else:
            return str(obj)

    return hashlib.sha256(bytes(as_str(obj), 'utf-8')).hexdigest()


def run_task(coro, callback):
    """
    Run awaitable in background task and return immediately

    This method should be used to call coroutine functions and other awaitables
    in a synchronous context.

    The returned task must be collected (e.g. in a :class:`list`) and awaited or
    cancelled eventually.

    If the returned task is cancelled, the :class:`asyncio.CancelledError` is
    silently ignored and `callback` is called with `None`.

    :param coro: Any awaitable object
    :param callback: Callable that is called with exception or return value of
        `coro`

    :return: :class:`asyncio.Task` instance
    """

    @functools.wraps(coro)
    async def wrap_coro(coro):
        try:
            return await asyncio.ensure_future(coro)
        except asyncio.CancelledError:
            pass
        except Exception as e:
            raise e

    def handle_task_done(task):
        try:
            result = task.result()
        except asyncio.CancelledError:
            # I don't know how to reliably reproduce or unittest this, but
            # CancelledError was raised here.
            pass
        except Exception as e:
            if callback:
                callback(e)
            else:
                raise
        else:
            if callback:
                callback(result)

    task = asyncio.ensure_future(wrap_coro(coro))
    task.add_done_callback(handle_task_done)
    return task


async def run_async(function, *args, **kwargs):
    """
    Run synchronous `function` asynchronously in a thread

    See :meth:`asyncio.BaseEventLoop.run_in_executor`.
    """
    loop = get_aioloop()
    wrapped = functools.partial(function, *args, **kwargs)
    return await loop.run_in_executor(None, wrapped)


# We must import these here to prevent circular imports
from . import (  # noqa: E402 isort:skip
    argtypes,
    browser,
    btclient,
    configfiles,
    country,
    daemon,
    fs,
    html,
    http,
    image,
    imghosts,
    predbs,
    release,
    signal,
    string,
    subproc,
    subtitle,
    timestamp,
    torrent,
    types,
    update,
    video,
    webdbs,
)



import functools

from . import JobWidgetBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class PosterJobWidget(JobWidgetBase):
    def setup(self):
        self.job.signal.register('finding_poster', self.handle_finding_poster)
        self.job.signal.register('downloading_poster', self.handle_downloading_poster)
        self.job.signal.register('resizing_poster', self.handle_resizing_poster)
        self.job.signal.register('uploading_poster', self.handle_uploading_poster)
        self.job.signal.register('uploaded_poster', self.handle_uploaded_poster)

    def handle_finding_poster(self, id):
        self.job.info = f'Searching {id}'

    def handle_downloading_poster(self, url):
        self.job.info = f'Downloading {url}'

    def handle_resizing_poster(self, filepath):
        self.job.info = f'Resizing {filepath}'

    def handle_uploading_poster(self, filepath):
        self.job.info = f'Uploading {filepath}'

    def handle_uploaded_poster(self, url):
        self.invalidate()

    @functools.cached_property
    def runtime_widget(self):
        return None



import re
from unittest.mock import Mock, PropertyMock, call

import pytest

from upsies.jobs import dialog


@pytest.fixture
def make_ChoiceJob(tmp_path):
    def make_ChoiceJob(**kwargs):
        return dialog.ChoiceJob(home_directory=tmp_path, cache_directory=tmp_path, **kwargs)
    return make_ChoiceJob


def test_name_property(make_ChoiceJob):
    job = make_ChoiceJob(name='foo', label='Foo', options=('1', '2', '3'))
    assert job.name == 'foo'


def test_label_property(make_ChoiceJob):
    job = make_ChoiceJob(name='foo', label='Foo', options=('1', '2', '3'))
    assert job.label == 'Foo'


def test_options_getter(make_ChoiceJob):
    job = make_ChoiceJob(name='foo', label='Foo', options=('1', '2', '3'))
    assert job.options is job._options
    assert job._options == [('1', '1'), ('2', '2'), ('3', '3')]

def test_options_setter_with_strings(make_ChoiceJob):
    job = make_ChoiceJob(name='foo', label='Foo', options=('1', '2', '3'))
    job.options = ('a', 'b', 'c')
    assert job._options == [('a', 'a'), ('b', 'b'), ('c', 'c')]

@pytest.mark.parametrize('invalid_sequence', (['b'], ['b', 2, 'foo']))
def test_options_setter_with_invalid_sequence(invalid_sequence, make_ChoiceJob):
    job = make_ChoiceJob(name='foo', label='Foo', options=('1', '2', '3'))
    with pytest.raises(ValueError, match=(
            r'^Option must be 2-tuple, '
            rf'not {re.escape(str(invalid_sequence))}$'
    )):
        job.options = (['a', 1], invalid_sequence, ['c', 3])
    assert job._options == [('1', '1'), ('2', '2'), ('3', '3')]

def test_options_setter_with_valid_sequence(make_ChoiceJob):
    job = make_ChoiceJob(name='foo', label='Foo', options=('1', '2', '3'))
    job.options = (['a', 1], ['b', 2], ['c', 3])
    assert job._options == [('a', 1), ('b', 2), ('c', 3)]

def test_options_setter_with_invalid_option(make_ChoiceJob):
    job = make_ChoiceJob(name='foo', label='Foo', options=('1', '2', '3'))
    with pytest.raises(ValueError, match=r'^Option must be 2-tuple, not None$'):
        job.options = (['a', 1], None, ['c', 3])
    assert job._options == [('1', '1'), ('2', '2'), ('3', '3')]

def test_options_setter_with_too_few_options(make_ChoiceJob):
    job = make_ChoiceJob(name='foo', label='Foo', options=('1', '2', '3'))
    with pytest.raises(ValueError, match=r"^There must be at least 2 options: \['a'\]$"):
        job.options = ['a']
    assert job._options == [('1', '1'), ('2', '2'), ('3', '3')]

@pytest.mark.parametrize('focused_option', (('2', 2), '2', 2))
def test_options_setter_preserves_focus_if_possible(focused_option, make_ChoiceJob):
    job = make_ChoiceJob(name='foo', label='Foo', options=(('1', 1), ('2', 2), ('3', 3)))
    job.focused = ('2', 2)
    assert job.focused == ('2', 2)
    job.options = (['a', 0], ['b', 1], ['c', 2], ['d', 3], ['e', 4])
    assert job.focused == ('c', 2)

@pytest.mark.parametrize('focused_option', (('2', 2), '2', 2))
def test_options_setter_defaults_focus_to_first_option(focused_option, make_ChoiceJob):
    job = make_ChoiceJob(name='foo', label='Foo', options=(('1', 1), ('2', 2), ('3', 3)))
    job.focused = ('3', 3)
    assert job.focused == ('3', 3)
    job.options = (['0', 0], ['1', 1], ['2', 2])
    assert job.focused == ('0', 0)

def test_options_setter_emits_dialog_updated_signal(make_ChoiceJob):
    job = make_ChoiceJob(name='foo', label='Foo', options=(('1', 1), ('2', 2), ('3', 3)))
    cb = Mock()
    job.signal.register('dialog_updated', cb)
    job.options = (['0', 0], ['1', 1], ['2', 2])
    assert cb.call_args_list == [call(job)]

def test_options_change_emits_dialog_updated_signal(make_ChoiceJob):
    job = make_ChoiceJob(name='foo', label='Foo', options=(('1', 1), ('2', 2), ('3', 3)))
    cb = Mock()
    job.signal.register('dialog_updated', cb)
    job.options.append(['3', 3])
    assert cb.call_args_list == [call(job)]


@pytest.mark.parametrize(
    argnames='options, thing, exp_result',
    argvalues=(
        # None
        ([('Foo', 10), ('Bar', 20), ('Baz', 30)], None, None),

        # option
        ([('Foo', 10), ('Bar', 20), ('Baz', 30)], ('Foo', 10), 0),
        ([('Foo', 10), ('Bar', 20), ('Baz', 30)], ('Bar', 20), 1),
        ([('Foo', 10), ('Bar', 20), ('Baz', 30)], ('Baz', 30), 2),

        # Index
        ([('Foo', 10), ('Bar', 20), ('Baz', 30)], -1, 0),
        ([('Foo', 10), ('Bar', 20), ('Baz', 30)], 0, 0),
        ([('Foo', 10), ('Bar', 20), ('Baz', 30)], 1, 1),
        ([('Foo', 10), ('Bar', 20), ('Baz', 30)], 2, 2),
        ([('Foo', 10), ('Bar', 20), ('Baz', 30)], 3, 2),

        # Label
        ([('Foo', 10), ('Bar', 20), ('Baz', 30)], 'Foo', 0),
        ([('Foo', 10), ('Bar', 20), ('Baz', 30)], 'Bar', 1),
        ([('Foo', 10), ('Bar', 20), ('Baz', 30)], 'Baz', 2),

        # Value
        ([('Foo', 1.1), ('Bar', 2.2), ('Baz', 3.3)], 1.1, 0),
        ([('Foo', 1.1), ('Bar', 2.2), ('Baz', 3.3)], 2.2, 1),
        ([('Foo', 1.1), ('Bar', 2.2), ('Baz', 3.3)], 3.3, 2),

        # Regular expression match against label
        ([('Foo', 1.1), ('Bar', 2.2), ('Baz', 3.3)], re.compile(r'(?i:foo)'), 0),
        ([('Foo', 1.1), ('Bar', 2.2), ('Baz', 3.3)], re.compile(r'r$'), 1),
        ([('Foo', 1.1), ('Bar', 2.2), ('Baz', 3.3)], re.compile(r'^[Bb]az$'), 2),

        # Regular expression match against value
        ([('Foo', 1.1), ('Bar', 2.2), ('Baz', 3.3)], re.compile(r'\d\.1'), 0),
        ([('Foo', 1.1), ('Bar', 2.2), ('Baz', 3.3)], re.compile(r'2+$'), 1),
        ([('Foo', 1.1), ('Bar', 2.2), ('Baz', 3.3)], re.compile(r'3\.?3'), 2),

        # No matching option found
        ([('Foo', 1.1), ('Bar', 2.2), ('Baz', 3.3)], 'foo', ValueError("No such option: 'foo'")),
    ),
    ids=lambda v: str(v),
)
def test_get_index(options, thing, exp_result, make_ChoiceJob):
    job = make_ChoiceJob(name='foo', label='Foo', options=options)

    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            job.get_index(thing)
    else:
        index = job.get_index(thing)
        assert index == exp_result

@pytest.mark.parametrize(
    argnames='options, index, focused_index, exp_return_value',
    argvalues=(
        ([('a', 1), ('b', 2), ('c', 3)], 0, 2, ('a', 1)),
        ([('a', 1), ('b', 2), ('c', 3)], 1, 1, ('b', 2)),
        ([('a', 1), ('b', 2), ('c', 3)], 2, 0, ('c', 3)),
        ([('a', 1), ('b', 2), ('c', 3)], None, 2, ('c', 3)),
        ([('a', 1), ('b', 2), ('c', 3)], None, 1, ('b', 2)),
        ([('a', 1), ('b', 2), ('c', 3)], None, 0, ('a', 1)),
    ),
    ids=lambda v: repr(v),
)
def test_get_option(options, index, focused_index, exp_return_value, make_ChoiceJob, mocker):
    job = make_ChoiceJob(name='foo', label='Foo', options=options)
    mocker.patch.object(job, 'get_index', return_value=index)
    mocker.patch.object(type(job), 'focused_index', PropertyMock(return_value=focused_index))

    return_value = job.get_option('something')
    assert return_value == exp_return_value
    assert job.get_index.call_args_list == [call('something')]


def test_focused_index(make_ChoiceJob):
    job = make_ChoiceJob(name='foo', label='Foo', options=('1', '2', '3'))
    assert job.focused_index is job._focused_index
    delattr(job, '_focused_index')
    assert job.focused_index is None


@pytest.mark.parametrize(
    argnames='options, focused_index, exp_focused',
    argvalues=(
        ((('1', 1), ('2', 2), ('3', 3)), None, None),
        ((('1', 1), ('2', 2), ('3', 3)), 0, ('1', 1)),
        ((('1', 1), ('2', 2), ('3', 3)), 1, ('2', 2)),
        ((('1', 1), ('2', 2), ('3', 3)), 2, ('3', 3)),
    ),
)
def test_focused_getter(options, focused_index, exp_focused, make_ChoiceJob, mocker):
    job = make_ChoiceJob(name='foo', label='Foo', options=options)
    mocker.patch.object(type(job), 'focused_index', PropertyMock(return_value=focused_index))
    assert job.focused == exp_focused

@pytest.mark.parametrize(
    argnames='get_index_return_value, exp_focused_index',
    argvalues=(
        (None, 0),
        (0, 0),
        (123, 123),
    ),
)
def test_focused_setter(get_index_return_value, exp_focused_index, make_ChoiceJob, mocker):
    job = make_ChoiceJob(name='foo', label='Foo', options=('1', '2', '3'))

    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(job, 'get_index', return_value=get_index_return_value), 'get_index')
    mocks.attach_mock(mocker.patch.object(job.signal, 'emit'), 'emit')

    job.focused = 'something'
    assert job.focused_index == exp_focused_index
    assert mocks.mock_calls == [
        call.get_index('something'),
        call.emit('dialog_updated', job),
    ]


def test_autodetected_index(make_ChoiceJob):
    job = make_ChoiceJob(name='foo', label='Foo', options=('1', '2', '3'))
    assert job.autodetected_index is job._autodetected_index
    delattr(job, '_autodetected_index')
    assert job.autodetected_index is None


@pytest.mark.parametrize(
    argnames='options, autodetected_index, exp_autodetected',
    argvalues=(
        ((('1', 1), ('2', 2), ('3', 3)), None, None),
        ((('1', 1), ('2', 2), ('3', 3)), 0, ('1', 1)),
        ((('1', 1), ('2', 2), ('3', 3)), 1, ('2', 2)),
        ((('1', 1), ('2', 2), ('3', 3)), 2, ('3', 3)),
    ),
)
def test_autodetected_getter(options, autodetected_index, exp_autodetected, make_ChoiceJob, mocker):
    job = make_ChoiceJob(name='foo', label='Foo', options=options)
    mocker.patch.object(type(job), 'autodetected_index', PropertyMock(return_value=autodetected_index))
    assert job.autodetected == exp_autodetected

@pytest.mark.parametrize(
    argnames='get_index_return_value, exp_autodetected_index',
    argvalues=(
        (None, None),
        (0, 0),
        (123, 123),
    ),
)
def test_autodetected_setter(get_index_return_value, exp_autodetected_index, make_ChoiceJob, mocker):
    job = make_ChoiceJob(name='foo', label='Foo', options=('1', '2', '3'))

    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(job, 'get_index', return_value=get_index_return_value), 'get_index')
    mocks.attach_mock(mocker.patch.object(job.signal, 'emit'), 'emit')

    job.autodetected = 'something'
    assert job.autodetected_index == exp_autodetected_index
    assert mocks.mock_calls == [
        call.get_index('something'),
        call.emit('dialog_updated', job),
    ]


def test_choice_property(make_ChoiceJob):
    job = make_ChoiceJob(name='foo', label='Foo', options=(('a', 1), ('b', 2), ('c', 3)))
    assert job.choice is None
    job._choice = 'mychoice'
    assert job.choice is job._choice


@pytest.mark.parametrize(
    argnames='choice_attribute, choice_return_value, exp_result',
    argvalues=(
        (None, ('asdf', 123), 123),
        ('mychoice', None, RuntimeError('foo: Choice was already made: mychoice')),
    ),
    ids=lambda v: str(v),
)
def test_set_choice(choice_attribute, choice_return_value, exp_result, make_ChoiceJob, mocker):
    job = make_ChoiceJob(name='foo', label='Foo', options=('ignored', 'options'))
    if choice_attribute is not None:
        job._choice = choice_attribute
    mocker.patch.object(job, 'get_option', return_value=choice_return_value)

    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            job._set_choice('something')
    else:
        job._set_choice('something')
        assert job._choice == exp_result
        assert job.get_option.call_args_list == [call('something')]


def test_make_choice(make_ChoiceJob, mocker):
    job = make_ChoiceJob(name='foo', label='Foo', options=('ignored', 'options'))
    mocker.patch.object(job, 'get_option', return_value=('asdf', 123))

    job.make_choice('something')
    assert job.output == ('asdf',)
    assert job.choice == 123
    assert job.is_finished


def test_set_label(make_ChoiceJob):
    job = make_ChoiceJob(name='foo', label='Foo', options=(('a', 1), ('b', 2), ('c', 3)))
    cb = Mock()
    job.signal.register('dialog_updated', cb)

    job.set_label('a', 'the a')
    assert job.options == [('the a', 1), ('b', 2), ('c', 3)]
    assert cb.call_args_list == [call(job)]

    job.set_label(2, 'the b')
    assert job.options == [('the a', 1), ('the b', 2), ('c', 3)]
    assert cb.call_args_list == [call(job), call(job)]

    job.set_label(('c', 3), 'the c')
    assert job.options == [('the a', 1), ('the b', 2), ('the c', 3)]
    assert cb.call_args_list == [call(job), call(job), call(job)]

    job.set_label('d', 'the d')
    assert job.options == [('the a', 1), ('the b', 2), ('the c', 3)]
    assert cb.call_args_list == [call(job), call(job), call(job)]


@pytest.mark.parametrize(
    argnames='autodetected, focused, exp_autodetected, exp_focused',
    argvalues=(
        (None, None, None, ('a', 'a')),
        ('b', None, ('b', 'b'), ('b', 'b')),
        ('b', 'c', ('b', 'b'), ('c', 'c')),
    ),
    ids=lambda v: repr(v),
)
def test_initialize_sets_properties(autodetected, focused, exp_autodetected, exp_focused, make_ChoiceJob, mocker):
    job = make_ChoiceJob(name='foo', label='Foo', options=('a', 'b', 'c'),
                         autodetected=autodetected, focused=focused)
    assert job.options == [('a', 'a'), ('b', 'b'), ('c', 'c')]
    assert job.autodetected == exp_autodetected
    assert job.focused == exp_focused

def test_initialize_sets_choice_property_when_output_is_cached(make_ChoiceJob):
    job1 = make_ChoiceJob(name='foo', label='Foo', options=(('a', 1), ('b', 2), ('c', 3)), ignore_cache=False)
    job1.start()
    job1.make_choice('b')
    assert job1.is_finished
    assert job1.output == ('b',)
    assert job1.choice == 2

    # Running the same job again should finish it immediately and set `choice`
    # as well as `output`
    job2 = make_ChoiceJob(name='foo', label='Foo', options=(('a', 1), ('b', 2), ('c', 3)), ignore_cache=False)
    job2.start()
    assert job2.is_finished
    assert job2.output == ('b',)
    assert job2.choice == 2


@pytest.mark.parametrize(
    argnames='start_starts, start_finishes, exp_autodetected, exp_focused, mocks, exp_calls',
    argvalues=(
        pytest.param(
            True, True, None, None,
            Mock(
                autodetect_and_finish=Mock(return_value='made choice and finished'),
                autodetect=Mock(return_value='made choice'),
            ),
            [
                ('JobBase_start', (), {}),
            ],
            id='super().start() finishes because it reads output from cache',
        ),
        pytest.param(
            False, False, None, None,
            Mock(
                autodetect_and_finish=Mock(return_value='made choice and finished'),
                autodetect=Mock(return_value='made choice'),
            ),
            [
                ('JobBase_start', (), {}),
            ],
            id='super().start() does not start (e.g. because it is not enabled)',
        ),
        pytest.param(
            True, False, None, None,
            Mock(
                autodetect_and_finish=None,
                autodetect=None,
            ),
            [
                ('JobBase_start', (), {}),
            ],
            id='No autodetection',
        ),
        pytest.param(
            True, False, ('c', 'c'), ('c', 'c'),
            Mock(
                autodetect_and_finish=Mock(return_value='c'),
                autodetect=None,
            ),
            [
                ('JobBase_start', (), {}),
                ('autodetect_and_finish', ('<job>',), {}),
                ('make_choice', (('c', 'c'),), {}),
            ],
            id='autodetect_and_finish succeeds',
        ),
        pytest.param(
            True, False, None, None,
            Mock(
                autodetect_and_finish=Mock(return_value=None),
                autodetect=None,
            ),
            [
                ('JobBase_start', (), {}),
                ('autodetect_and_finish', ('<job>',), {}),
            ],
            id='autodetect_and_finish fails',
        ),
        pytest.param(
            True, False, ('a', 'a'), ('a', 'a'),
            Mock(
                autodetect_and_finish=None,
                autodetect=Mock(return_value='a'),
            ),
            [
                ('JobBase_start', (), {}),
                ('autodetect', ('<job>',), {}),
            ],
            id='autodetect succeeds',
        ),
        pytest.param(
            True, False, None, None,
            Mock(
                autodetect_and_finish=None,
                autodetect=Mock(return_value=None),
            ),
            [
                ('JobBase_start', (), {}),
                ('autodetect', ('<job>',), {}),
            ],
            id='autodetect fails',
        ),
    ),
    ids=lambda v: repr(v),
)
def test_start(start_starts, start_finishes, exp_autodetected, exp_focused, mocks, exp_calls, make_ChoiceJob, mocker):
    job = make_ChoiceJob(name='foo', label='Foo',
                         options=('a', 'b', 'c'),
                         focused='b',
                         autodetect=mocks.autodetect,
                         autodetect_and_finish=mocks.autodetect_and_finish)
    mocks.attach_mock(mocker.patch.object(job, 'make_choice'), 'make_choice')

    def JobBase_start():
        if start_finishes:
            job.finish()
        elif start_starts:
            job._is_started = True

    mocks.attach_mock(mocker.patch('upsies.jobs.base.JobBase.start', side_effect=JobBase_start), 'JobBase_start')

    exp_focused = job.focused if exp_focused is None else exp_focused

    job.start()

    if start_finishes:
        assert job.is_finished

    assert job.autodetected == exp_autodetected
    assert job.focused == exp_focused

    # Because the `job` instance is not available during parameter creation, we
    # have to replace placeholders
    exp_actual_calls = []
    for c in exp_calls:
        print(c)
        name = getattr(call, c[0])
        args = [
            job if arg == '<job>' else arg
            for arg in c[1]
        ]
        kwargs = c[2]
        exp_actual_calls.append(name(*args, **kwargs))
        print(exp_actual_calls[-1])

    assert mocks.mock_calls == exp_actual_calls



from unittest.mock import Mock, call

import pytest

from upsies.jobs.torrent import CopyTorrentJob


@pytest.fixture
async def make_CopyTorrentJob(tmp_path):
    def make_CopyTorrentJob(destination, enqueue=()):
        return CopyTorrentJob(
            home_directory=tmp_path,
            cache_directory=tmp_path,
            ignore_cache=False,
            destination=destination,
            enqueue=enqueue,
        )
    return make_CopyTorrentJob


def test_cache_id(make_CopyTorrentJob, tmp_path):
    job = make_CopyTorrentJob(destination=tmp_path)
    assert job.cache_id is None


@pytest.mark.asyncio
async def test_handle_input_call_order(make_CopyTorrentJob, mocker, tmp_path):
    job = make_CopyTorrentJob(destination=tmp_path)

    filepath = tmp_path / 'foo.torrent'
    filepath.write_text('hello')

    mocks = Mock()
    job.signal.register('copying', mocks.copying)
    job.signal.register('copied', mocks.copied)
    copy2_mock = mocker.patch('shutil.copy2', return_value=str(tmp_path / filepath.name))
    mocks.attach_mock(copy2_mock, 'copy2')

    await job.handle_input(str(filepath))
    assert mocks.mock_calls == [
        call.copying(str(filepath)),
        call.copy2(str(filepath), str(tmp_path)),
        call.copied(copy2_mock.return_value),
    ]

@pytest.mark.asyncio
async def test_handle_input_reports_nonexisting_file(make_CopyTorrentJob, mocker, tmp_path):
    job = make_CopyTorrentJob(destination='foo')

    mocks = Mock()
    job.signal.register('copying', mocks.copying)
    job.signal.register('copied', mocks.copied)
    copy2_mock = mocker.patch('shutil.copy2')
    mocks.attach_mock(copy2_mock, 'copy2')

    await job.handle_input(tmp_path / 'foo')
    assert job.output == ()
    assert job.errors == (f'{tmp_path / "foo"}: No such file',)
    assert mocks.mock_calls == []

@pytest.mark.asyncio
async def test_copy_too_large_file(make_CopyTorrentJob, mocker, tmp_path):
    job = make_CopyTorrentJob(destination='foo')

    filepath = tmp_path / 'foo.torrent'
    with open(filepath, 'wb') as f:
        f.truncate(CopyTorrentJob.MAX_FILE_SIZE + 1)  # Sparse file

    mocks = Mock()
    job.signal.register('copying', mocks.copying)
    job.signal.register('copied', mocks.copied)
    copy2_mock = mocker.patch('shutil.copy2')
    mocks.attach_mock(copy2_mock, 'copy2')

    await job.handle_input(filepath)
    assert job.output == ()
    assert job.errors == (f'{filepath}: File is too large',)
    assert mocks.mock_calls == []

@pytest.mark.asyncio
async def test_handle_input_sends_destination_path_on_success(make_CopyTorrentJob, mocker, tmp_path):
    job = make_CopyTorrentJob(destination='foo')

    filepath = tmp_path / 'foo.torrent'
    filepath.write_bytes(b'content')

    mocker.patch('shutil.copy2', return_value='destination/path')

    await job.handle_input(filepath)
    assert job.output == ('destination/path',)
    assert job.errors == ()

@pytest.mark.asyncio
async def test_handle_input_sends_source_path_on_failure(make_CopyTorrentJob, mocker, tmp_path):
    job = make_CopyTorrentJob(destination='foo')

    filepath = tmp_path / 'foo.torrent'
    filepath.write_bytes(b'content')

    mocker.patch('shutil.copy2', side_effect=PermissionError('Permission denied'))

    await job.handle_input(filepath)
    assert job.output == ()
    assert job.errors == (f'Failed to copy {filepath} to {job._destination}: Permission denied',)

@pytest.mark.asyncio
async def test_handle_input_sets_info_property_on_success(make_CopyTorrentJob, mocker, tmp_path):
    job = make_CopyTorrentJob(destination='foo')

    filepath = tmp_path / 'foo.torrent'
    filepath.write_bytes(b'content')

    mocker.patch('shutil.copy2', return_value='destination/path')

    infos = [
        'Copying foo.torrent',
        '',
    ]

    def info_cb(_):
        assert job.info == infos.pop(0)

    job.signal.register('copying', info_cb)
    job.signal.register('copied', info_cb)
    job.signal.register('error', info_cb)
    job.signal.register('finished', info_cb)
    await job.handle_input(filepath)
    assert infos == []

@pytest.mark.asyncio
async def test_handle_input_sets_info_property_on_failure(make_CopyTorrentJob, mocker, tmp_path):
    job = make_CopyTorrentJob(destination='foo')

    filepath = tmp_path / 'foo.torrent'
    filepath.write_bytes(b'content')

    mocker.patch('shutil.copy2', side_effect=OSError('No'))

    infos = [
        'Copying foo.torrent',
        '',
        '',
    ]

    def info_cb(_):
        assert job.info == infos.pop(0)

    job.signal.register('copying', info_cb)
    job.signal.register('copied', info_cb)
    job.signal.register('error', info_cb)
    job.signal.register('finished', info_cb)
    await job.handle_input(filepath)
    assert infos == []



from base64 import b64decode

from ... import errors
from .. import http
from . import base

import logging  # isort:skip
_log = logging.getLogger(__name__)


class PredbovhApi(base.PredbApiBase):
    name = 'predbovh'
    label = 'PreDB.ovh'

    default_config = {}

    _url_base = b64decode('cHJlZGIub3Zo').decode('ascii')
    _search_url = f'https://{_url_base}/api/v1/'

    async def _search(self, query):
        q = self._get_q(query.keywords, query.group)
        return await self._request_all_pages(q)

    def _get_q(self, keywords, group):
        if group:
            keywords = list(keywords)
            keywords.extend(('@team', str(group).replace('@', r'\@').strip()))
        kws = (str(kw).lower().strip() for kw in keywords)
        return ' '.join(kw for kw in kws if kw)

    async def _request_all_pages(self, q):
        combined_results = []

        # We can request 30 pages per minute before we get an error
        for page in range(1, 31):
            results, next_page = await self._request_page(q, page)
            combined_results.extend(results)

            # Negative next page means last page
            if next_page < 0:
                break

        return combined_results

    async def _request_page(self, q, page):
        params = {
            'q': q,
            'count': 100,
            'page': page,
        }
        _log.debug('%s search: %r, %r', self.label, self._search_url, params)
        response = (await http.get(self._search_url, params=params, cache=True)).json()

        # Report API error or return list of release names
        if response['status'] != 'success':
            raise errors.RequestError(f'{self.label}: {response["message"]}')
        else:
            # Extract release names
            results = tuple(result['name'] for result in response['data']['rows'])

            # Is there another page of results?
            if len(results) >= response['data']['reqCount']:
                next_page = page + 1
            else:
                next_page = -1

            return results, next_page

    async def _release_files(self, release_name):
        raise NotImplementedError()



from unittest.mock import Mock

import pytest

from upsies.uis.tui import utils


@pytest.mark.parametrize(
    argnames='stdin_isatty, stdout_isatty, stderr_isatty',
    argvalues=(
        (False, False, False),
        (False, False, True),
        (False, True, False),
        (False, True, True),
        (True, False, False),
        (True, False, True),
        (True, True, False),
        (True, True, True),
    ),
)
@pytest.mark.parametrize(
    argnames='stdin, stdout, stderr',
    argvalues=(
        (None, None, None),
        (None, None, Mock()),
        (None, None, Mock()),
        (None, Mock(), None),
        (None, Mock(), Mock()),
        (Mock(), None, None),
        (Mock(), None, Mock()),
        (Mock(), Mock(), None),
        (Mock(), Mock(), Mock()),
    ),
)
def test_is_tty(stdin, stdout, stderr, stdin_isatty, stdout_isatty, stderr_isatty, mocker):
    if stdin:
        stdin.isatty = lambda: stdin_isatty
    if stdout:
        stdout.isatty = lambda: stdout_isatty
    if stderr:
        stderr.isatty = lambda: stderr_isatty
    mocker.patch('sys.stdin', stdin)
    mocker.patch('sys.stdout', stdout)
    mocker.patch('sys.stderr', stderr)

    if not stdin or not stdin.isatty():
        exp_is_tty = False
    elif (stdout and stdout.isatty()) or (stderr and stderr.isatty()):
        exp_is_tty = True
    else:
        exp_is_tty = False

    assert utils.is_tty() is exp_is_tty
